@@ -445,9 +445,11 @@ namespace ampspy
       void connectionStateChanged(State newState_)
       {
         LOCKGIL;
-        PyObject_CallObject(_listener, Py_BuildValue("(O)",
+        PyObject* args = Py_BuildValue("(O)",
                             (newState_ == AMPS::ConnectionStateListener::Disconnected ||
-                             newState_ == AMPS::ConnectionStateListener::Shutdown) ? Py_False : Py_True));
+                             newState_ == AMPS::ConnectionStateListener::Shutdown) ? Py_False : Py_True);
+        Py_XDECREF(PyObject_CallObject(_listener, args));
+        Py_DECREF(args);
       }
     };
 
@@ -468,8 +470,11 @@ namespace ampspy
       void connectionStateChanged(State newState_)
       {
         LOCKGIL;
-        PyObject_CallObject(_listener, Py_BuildValue("(O)",
-                            PyInt_FromLong((long)newState_)));
+        PyObject* pyState = PyInt_FromLong((long)newState_);
+        PyObject* args = Py_BuildValue("(O)", pyState);
+        Py_XDECREF(PyObject_CallObject(_listener, args));
+        Py_DECREF(args);
+        Py_DECREF(pyState);
       }
     };
 
@@ -1654,7 +1659,7 @@ namespace ampspy
       obj* c = (obj*) vp;
 
       PyObject* args = Py_BuildValue("(O)", c);
-      PyObject_Call(c->disconnectHandler, args, (PyObject*)NULL);
+      Py_XDECREF(PyObject_Call(c->disconnectHandler, args, (PyObject*)NULL));
       Py_DECREF(args);
       if (PyErr_Occurred())
       {
@@ -2085,14 +2090,14 @@ namespace ampspy
       {
         LOCKGIL;
 #if PY_MAJOR_VERSION >= 3
-        PyObject_CallObject(pSelf->_handler,
-                            Py_BuildValue("(y#O)", data_, len_,
-                                          (direction_ ? Py_True : Py_False)));
+         PyObject* args = Py_BuildValue("(y#O)", data_, len_,
+                                        (direction_ ? Py_True : Py_False));
 #else
-        PyObject_CallObject(pSelf->_handler,
-                            Py_BuildValue("(s#O)", data_, len_,
-                                          (direction_ ? Py_True : Py_False)));
+        PyObject* args = Py_BuildValue("(s#O)", data_, len_,
+                                       (direction_ ? Py_True : Py_False));
 #endif
+        Py_XDECREF(PyObject_CallObject(pSelf->_handler, args));
+        Py_DECREF(args);
       }
       catch (...) { } // -V565 this is a C callback function, so exceptions must not bubble up.
     }
@@ -2145,9 +2150,10 @@ namespace ampspy
 #endif
           PyObject* data = PyMemoryView_FromMemory((char*)start, (Py_ssize_t)sz,
                                                    PyBUF_WRITE);
-          PyObject_CallObject(pSelf->_handler,
-                              Py_BuildValue("(OO)", data,
-                                            (direction_ ? Py_True : Py_False)));
+          PyObject* args = Py_BuildValue("(OO)", data,
+                                         (direction_ ? Py_True : Py_False));
+          Py_XDECREF(PyObject_CallObject(pSelf->_handler, args));
+          Py_DECREF(args);
 #else
           Py_ssize_t shape[] = { (Py_ssize_t)sz };
           Py_buffer buf;
@@ -2165,8 +2171,11 @@ namespace ampspy
           buf.smalltable[1] = 1;
           buf.internal = NULL;
           PyObject* data = PyMemoryView_FromBuffer(&buf);
-          PyObject_CallObject(pSelf->_handler, Py_BuildValue("(OO)", data,
-                              (direction_ ? Py_True : Py_False)));
+          PyObject* args = Py_BuildValue("(OO)", data,
+                              (direction_ ? Py_True : Py_False));
+          Py_XDECREF(PyObject_CallObject(pSelf->_handler, args));
+          Py_DECREF(data);
+          Py_DECREF(args);
 #endif
         }
         catch (...) { } // -V565 this is a C callback function, so exceptions must not bubble up.
@@ -2260,7 +2269,7 @@ namespace ampspy
       try
       {
         LOCKGIL;
-        PyObject_CallObject(callable, NULL);
+        Py_XDECREF(PyObject_CallObject(callable, NULL));
         return AMPS_E_OK;
       }
       catch (...)
@@ -2482,6 +2491,16 @@ namespace ampspy
       CALL_RETURN_SELF(((AMPS::Client*)self->pClient)->clearHttpPreflightHeaders());
     }
 
+    static PyObject* set_publish_batching(obj* self, PyObject* args)
+    {
+      unsigned long batchSizeBytes, batchTimeoutMillis;
+      if (!PyArg_ParseTuple(args, "kk", &batchSizeBytes, &batchTimeoutMillis))
+      {
+        return NULL;
+      }
+      CALL_RETURN_SELF(((AMPS::Client*)self->pClient)->setPublishBatching((amps_uint64_t)batchSizeBytes, (amps_uint64_t)batchTimeoutMillis));
+    }
+
 ///
 /// This method is used by python internals for garbage collection.
 ///
@@ -2693,6 +2712,8 @@ namespace ampspy
                  add_http_preflight_header_key_value_doc)
       .addMethod("clear_http_preflight_headers", clear_http_preflight_headers,
                  clear_http_preflight_headers_doc)
+      .addMethod("set_publish_batching", set_publish_batching,
+                 set_publish_batching_doc)
       .addGetter("_as_parameter_", (void*)get_as_parameter,
                  "Underlying Client pointer for use with foreign functions and the ctypes module.")
       .createType()
