@@ -52,6 +52,7 @@ An example of a Python client publishing a JSON message is listed below::
 
 **Constructor Arguments:**
 
+
 :param name: The unique name for this client. AMPS does not enforce
              specific restrictions on the character set used, however some protocols
              (for example, XML) may not allow specific characters. 60East recommends
@@ -83,6 +84,7 @@ Creates a new :class:`Message` appropriate for this client.
 
 This function should be called rarely, since it does allocate a handful of small objects.
 
+
 :returns: A new *Message* instance.
 )docstring";
 
@@ -92,6 +94,7 @@ Execute the provided command and process responses using a ``MessageStream``. Th
 a ``Message`` based on the provided ``Command``, sends the ``Message`` to AMPS, and returns a
 ``MessageStream`` that can be iterated on to process messages returned in response to the command.
 
+
 :type command: AMPS.Command
 :param command: The command to execute.
 :returns: A ``MessageStream`` to iterate over.
@@ -107,6 +110,7 @@ messages on the calling thread, the AMPS Python client runs the handler directly
 When the provided handler is not ``None``, this function blocks until AMPS acknowledges that the command has
 been processed. The results of the command after that acknowledgment are provided to the handler.
 
+
 :type on_message: A function or other callable object.
 :type command: AMPS.Command
 :param command: The command to execute.
@@ -118,6 +122,7 @@ been processed. The results of the command after that acknowledgment are provide
 static const char* get_unpersisted_count_doc = R"docstring(get_unpersisted_count()
 Gets the count of unpersisted publishes.
 
+
 :returns: The count of unpersisted publishes.
 )docstring";
 
@@ -125,6 +130,7 @@ Gets the count of unpersisted publishes.
 static const char* get_server_version_doc = R"docstring(get_server_version()
 Gets the connected server's version as a numeric value.
 
+
 :returns: The version converted to a number. For example, server version 3.8.1.7 would be 3080107.
 )docstring";
 
@@ -132,6 +138,7 @@ Gets the connected server's version as a numeric value.
 static const char* get_server_version_info_doc = R"docstring(get_server_version_info()
 Gets the connected server's version as a :class:`VersionInfo`.
 
+
 :returns: The version in a :class:`VersionInfo`.
 )docstring";
 
@@ -140,6 +147,7 @@ static const char* convert_version_to_number_doc = R"docstring(convert_version_t
 Converts the provided version string to a number using 2 digits for each dot.
 A value such as 4.0.0.0 will become 4000000 while 3.8.1.7 will return 3080107.
 
+
 :param version: The version string to convert.
 :type version: str
 :returns: The numeric value for the version string.
@@ -164,6 +172,7 @@ static const char* bookmark_subscribe_doc = R"docstring(bookmark_subscribe()
      messages received from the transaction log replay (if any) and then cuts over to live messages that match
      the subscription.
 
+
 :param on_message: The message handler to invoke with matching messages
                    (specified only in the case where async processing is used).
 :type on_message: :class:`MessageHandler`
@@ -225,6 +234,7 @@ Connects to the AMPS instance through the provided URI.
          regardless of ``userId`` or ``password``), you can omit the
          ``userId:password@`` string from the connect URI.
 
+
 :param uri: The URI used to connect to AMPS.
 :type uri: str
 :raises: :class:`ConnectionException`
@@ -254,6 +264,7 @@ To detect failure, install a failed write handler.
 If the client was created with a persistent store on construction,
 then the client will store before forwarding the message to AMPS.
 
+
 :param topic: The topic to publish the data to.
 :type topic: str
 :param data: The data to publish to the topic.
@@ -284,6 +295,7 @@ static const char* delta_subscribe_doc = R"docstring(delta_subscribe()
   2. When no message handler is provided, this method returns a message stream that can be iterated on to process 
      messages received that match the subscription and that contain the updated fields.
 
+
 :param on_message: The message handler to invoke with matching messages
                    (specified only in the case where async processing
                    is used).
@@ -343,6 +355,7 @@ Gets the string name hash of the :class:`Client` object.
 static const char* get_name_hash_value_doc = R"docstring(get_name_hash_value()
 Gets the numeric name hash of the :class:`Client` object.
 
+
 :returns: The name hash int.
 )docstring";
 
@@ -358,6 +371,7 @@ A legacy method name for :meth:`set_name`.
 static const char* set_name_doc = R"docstring(set_name(name)
 Sets the :class:`Client` name.
 
+
 :param name: The :class:`Client` name.
 )docstring";
 
@@ -365,6 +379,7 @@ Sets the :class:`Client` name.
 static const char* get_logon_correlation_data_doc = R"docstring(get_logon_correlation_data()
 Gets the data used to correlate the logon of this :class:`Client` to the server.
 
+
 :returns: The logon correlation data.
 )docstring";
 
@@ -372,6 +387,7 @@ Gets the data used to correlate the logon of this :class:`Client` to the server.
 static const char* set_logon_correlation_data_doc = R"docstring(set_logon_correlation_data(logon_correlation_data)
 Sets the data used to correlate the logon of this Client in the server.
 
+
 :param logon_correlation_data: The base64 data string to send with the logon.
 )docstring";
 
@@ -379,6 +395,7 @@ Sets the data used to correlate the logon of this Client in the server.
 static const char* logon_doc = R"docstring(logon(timeout=0,authenticator=AMPS.DefaultAuthenticator,options=None)
 Logs into AMPS with the parameters provided in the :meth:`connect` method.
 
+
 :param timeout: The maximum time to wait for the command to
                 receive a processed ack from AMPS for the logon command
                 (in milliseconds). ``0`` indicates to wait indefinitely.
@@ -397,6 +414,7 @@ Logs into AMPS with the parameters provided in the :meth:`connect` method.
 static const char* name_doc = R"docstring(name()
 The name of the :class:`Client` object.
 
+
 :returns: The :class:`Client` name.
 )docstring";
 
@@ -407,6 +425,7 @@ To detect failure, install a failed write handler. If the client was created wit
 on construction, then the client will store before forwarding the message to AMPS. If a :exc:`DisconnectedException`
 occurs, the message is still stored in the publish store.
 
+
 :param topic: The topic to publish to.
 :type topic: str
 :param data: The data to publish.
@@ -421,9 +440,11 @@ occurs, the message is still stored in the publish store.
 )docstring";
 
 
+
 static const char* send_doc = R"docstring(send(message,message_handler=None,timeout=None)
 Send a :class:`Message` to AMPS via the Transport used in the :class:`Client`.
 
+
 :param message: The message to send.
 :type message: :class:`Message`
 :param message_handler: The message handler that will receive messages
@@ -443,6 +464,7 @@ static const char* add_message_handler_doc = R"docstring(add_message_handler(com
 Add a message handler to the :class:`Client` to handle messages and the
 requested acknowledgments sent in response to the given ``command_id``.
 
+
 :param command_id: The command, query, or sub ID for messages and acks.
 :type command_id: str
 :param message_handler: The message handler that will receive messages
@@ -458,6 +480,7 @@ requested acknowledgments sent in response to the given ``command_id``.
 static const char* remove_message_handler_doc = R"docstring(remove_message_handler(command_id)
 Remove a message handler from the :class:`Client`.
 
+
 :param command_id: The command ID for the handler to remove.
 :type command_id: str
 )docstring";
@@ -510,6 +533,7 @@ disconnected from AMPS, the invoke method from the :class:`ClientDisconnectHandl
  .. deprecated:: 5.3.5.0
   Use :class:`HAClient` instead to get automatic disconnect handling.
 
+
 :param client_disconnect_handler: The disconnect handler.
 :type client_disconnect_handler: :class:`DisconnectHandler`
 )docstring";
@@ -518,6 +542,7 @@ disconnected from AMPS, the invoke method from the :class:`ClientDisconnectHandl
 static const char* set_exception_listener_doc = R"docstring(set_exception_listener(exception_listener)
 Sets the exception listener instance used for communicating absorbed exceptions.
 
+
 :param exception_listener: The exception listener instance to invoke for exceptions.
 :type exception_listener: Exception
 )docstring";
@@ -526,6 +551,7 @@ Sets the exception listener instance used for communicating absorbed exceptions.
 static const char* get_exception_listener_doc = R"docstring(get_exception_listener()
 Gets the exception listener callable set on self.
 
+
 :returns: The exception listener callable set on self, or None.
 )docstring";
 
@@ -549,6 +575,7 @@ asynchronous message processing, your message handler must process
 messages within the timeout interval, or risk being disconnected
 by the server.
 
+
 :param interval_seconds: The time between heartbeat messages being sent to AMPS.
 :type interval_seconds: int
 :param timeout_seconds: The maximum time to wait for AMPS to acknowledge the start of heartbeating (in seconds).
@@ -568,6 +595,7 @@ Used to stop a timer on an AMPS Server previously started for the client.
 
 .. deprecated:: 5.3.2.0
 
+
 :param handler: The handler to be invoked with the timer response.
 )docstring";
 
@@ -575,6 +603,7 @@ Used to stop a timer on an AMPS Server previously started for the client.
 static const char* set_last_chance_message_handler_doc = R"docstring(set_last_chance_message_handler(message_handler)
 Sets the :class:`MessageHandler` instance called when no other incoming message handler matches.
 
+
 :param message_handler: The message handler to invoke when no other incoming message handler matches.
 :type message_handler: :class:`MessageHandler`
 )docstring";
@@ -584,6 +613,7 @@ static const char* set_duplicate_message_handler_doc = R"docstring(set_duplicate
 Sets the :class:`MessageHandler` instance used for messages that arrive from AMPS that are deemed to be duplicates
 of previous messages, according to the local bookmark store.
 
+
 :param message_handler: The message handler to invoke for duplicate messages.
 :type message_handler: MessageHandler
 )docstring";
@@ -592,6 +622,7 @@ of previous messages, according to the local bookmark store.
 static const char* get_duplicate_message_handler_doc = R"docstring(get_duplicate_message_handler()
 Gets the message handler object set with :meth:`set_duplicate_message_handler`.
 
+
 :returns: The message handler object set with :meth:`set_duplicate_message_handler`.
 )docstring";
 
@@ -626,6 +657,7 @@ For example::
   finally:
       client.close()
 
+
 :param on_message: The message handler to invoke with matching
                    messages (specified only in the case where async
                    processing is used).
@@ -703,6 +735,7 @@ static const char* sow_and_delta_subscribe_doc = R"docstring(sow_and_delta_subsc
      beginning and end of the SOW results and then processes messages received that match the subscription and that
      contain the updated fields.
 
+
 :param on_message: The message handler to invoke with matching
                    messages (specified only in the case where async processing is used).
 :type on_message: :class:`MessageHandler`
@@ -757,6 +790,7 @@ static const char* sow_and_subscribe_doc = R"docstring(sow_and_subscribe()
      messages received from the SOW query, including the ``group_begin`` and ``group_end`` messages, that indicate the
      beginning and end of the SOW results and then processes messages received that match the subscription.
 
+
 :param on_message: The message handler to invoke with matching
                    messages (specified only in the case where async processing is used).
 :type on_message: :class:`MessageHandler`
@@ -814,6 +848,7 @@ For example, to delete all messages that match a filter::
   print("%s: %s" % (ackMessage.get_ack_type(), ackMessage.get_status()))
   ...
 
+
 :param on_message: The message handler to invoke with `stats` and `completed` acknowledgments
                    (specified only in the case where async processing is used).
 :type  on_message: :exc:`MessageHandler`
@@ -850,6 +885,7 @@ static const char* sow_delete_by_keys_doc = R"docstring(sow_delete_by_keys()
   2. When no message handler is provided, this method provides an acknowledgment message with the result of
      the delete command.
 
+
 :param on_message: The message handler to invoke with `stats` and `completed` acknowledgments
                    (specified only in the case where async processing is used).
 :type on_message: :exc:`MessageHandler`
@@ -899,6 +935,7 @@ that is already stored, for example::
   ackMessage = client.sow_delete_by_data('orders', data)
   del orders[orderId]
 
+
 :param on_message: The message handler to invoke with `stats` and `completed` acknowledgments
                    (specified only in the case where async processing is used).
 :type on_message: :exc:`MessageHandler`
@@ -932,6 +969,7 @@ static const char* subscribe_doc = R"docstring(subscribe()
   2. When no message handler is provided, this method returns a message stream that can be iterated on to process
      messages received that match the subscription.
 
+
 :param on_message: The message handler to invoke with matching
                    messages (specified only in the case where async processing is used).
 :type on_message: str
@@ -967,6 +1005,7 @@ Remove a subscription from AMPS.
     .. note::
       Using the keyword ``all`` with ``sub_id`` will remove all subscriptions to AMPS.
 
+
 :param sub_id: The subscription ID to remove.
 :type sub_id: str
 )docstring";
@@ -975,6 +1014,7 @@ Remove a subscription from AMPS.
 static const char* get_uri_doc = R"docstring(get_uri()
 Gets the URI string passed into the :class:`Client` during the :meth:`connect` invocation.
 
+
 :returns: The URI string.
 )docstring";
 
@@ -995,6 +1035,7 @@ delivered to the AMPS server. For example, an application might call
 One thing to note is that if AMPS is unavailable (HA Client), ``publish_flush``
 needs to wait for a connection to come back up, which may look like it's hanging.
 
+
 :param timeout: The maximum time to wait for the messages to be acknowledged
                 as persisted, or for the flush command to be acknowledged
                 by AMPS (in milliseconds). ``0`` indicates to wait indefinitely.
@@ -1008,6 +1049,7 @@ needs to wait for a connection to come back up, which may look like it's hanging
 static const char* flush_doc = R"docstring(flush(timeout=0)
 Another name for :meth:`publish_flush`.
 
+
 :param timeout: The maximum time to wait for the messages to be acknowledged
                 as persisted, or for the flush command to be acknowledged
                 by AMPS (in milliseconds). ``0`` indicates to wait indefinitely.
@@ -1018,6 +1060,7 @@ Another name for :meth:`publish_flush`.
 static const char* set_publish_store_doc = R"docstring(set_publish_store(publish_store)
 Sets a publish store on self.
 
+
 :param publish_store: A :class:`PublishStore` or :class:`MemoryPublishStore` instance.
 )docstring";
 
@@ -1025,6 +1068,7 @@ Sets a publish store on self.
 static const char* set_bookmark_store_doc = R"docstring(set_bookmark_store(bookmark_store)
 Sets a bookmark store on self.
 
+
 :param bookmark_store: A :class:`MMapBookmarkStore` or :class:`MemoryBookmarkStore` instance,
                       or a custom object that implements the required bookmark store methods.
 )docstring";
@@ -1042,6 +1086,7 @@ For example, you might implement a function like::
                    (message.get_topic(), message.get_data()[0:20])
      print(output)
 
+
 :param failedWriteHandler: A callable object to be invoked when AMPS indicates that a published message is not
                            written. This could be because a duplicate message already exists in the transaction
                            log, this client is not entitled to publish to the topic, the message failed to parse,
@@ -1054,6 +1099,7 @@ For example, you might implement a function like::
 static const char* add_connection_state_listener_doc = R"docstring(add_connection_state_listener(listener_callable)
 Sets a function to be called when this client connects or disconnects from AMPS.
 
+
 :type listener_callable: A python function or other callable that takes a single value.
 :param listener_callable: This function will be passed True if a connection is established, False if
                           a disconnect occurs. Notice that when a connection is established, the client
@@ -1065,6 +1111,7 @@ Sets a function to be called when this client connects or disconnects from AMPS.
 static const char* remove_connection_state_listener_doc = R"docstring(remove_connection_state_listener(listener_callable)
 Removes a listener function previously supplied to :meth:`add_connection_state_listener`.
 
+
 :type listener_callable: A python function or other callable that takes a single value.
 :param listener_callable: The function or callable to be removed.
 )docstring";
@@ -1073,6 +1120,7 @@ Removes a listener function previously supplied to :meth:`add_connection_state_l
 static const char* set_transport_filter_doc = R"docstring(set_transport_filter(transport_filter_callable)
 Sets a function to be called when this client sends or receives data.
 
+
 :type transport_filter_callable: A python function or other callable that takes two parameters: data and direction.
 :param transport_filter_callable: This function is passed a string ('data') containing the raw bytes sent or received.
                                   The 'direction' parameter is False if the data is being sent to the server, or True
@@ -1083,6 +1131,7 @@ Sets a function to be called when this client sends or receives data.
 static const char* set_thread_created_callback_doc = R"docstring(set_thread_created_callback(thread_created_callable)
 Sets a function to be called when this client creates a thread used to receive data from the server.
 
+
 :type thread_created_callable: A python function or other callable that takes no parameters.
 :param thread_created_callable: This function is called by the newly created thread.
 )docstring";
@@ -1091,6 +1140,7 @@ Sets a function to be called when this client creates a thread used to receive d
 static const char* ack_doc = R"docstring(ack(message,options=None) OR ack(topic,bookmark,options=None)
 Acknowledges a message queue message.
 
+
 :param message: An :class:`AMPS.Message` object to ack, OR
 :param topic: The topic of the message to ack.
 :param bookmark: The bookmark of the message to ack.
@@ -1101,6 +1151,7 @@ Acknowledges a message queue message.
 static const char* set_ack_batch_size_doc = R"docstring(set_ack_batch_size(batch_size)
 Sets the batch size used for batching message queue ack messages.
 
+
 :param batch_size: The number of ack messages to batch before sending.
 )docstring";
 
@@ -1108,6 +1159,7 @@ Sets the batch size used for batching message queue ack messages.
 static const char* get_ack_batch_size_doc = R"docstring(get_ack_batch_size()
 Gets the current batch size used for batching message queue acknowledgment messages.
 
+
 :returns: The current batch size used for batching message queue ack messages.
 )docstring";
 
@@ -1115,6 +1167,7 @@ Gets the current batch size used for batching message queue acknowledgment messa
 static const char* set_auto_ack_doc = R"docstring(set_auto_ack(enabled)
 Enables or disables auto-acknowledgment of message queue messages.
 
+
 :param enabled: True to enable auto-acknowledgment of message queue messages.
 :type enabled: Boolean
 )docstring";
@@ -1123,6 +1176,7 @@ Enables or disables auto-acknowledgment of message queue messages.
 static const char* get_auto_ack_doc = R"docstring(get_auto_ack()
 Called to check if auto-acknowledgment of message queue messages is enabled.
 
+
 :returns: True, if auto-acknowledgment of message queue messages is enabled.
 )docstring";
 
@@ -1130,6 +1184,7 @@ Called to check if auto-acknowledgment of message queue messages is enabled.
 static const char* get_ack_timeout_doc = R"docstring(get_ack_timeout()
 Gets the current time (milliseconds) before queued acknowledgment messages are sent.
 
+
 :returns: The current time (milliseconds) before queued ack messages are sent.
 )docstring";
 
@@ -1137,6 +1192,7 @@ Gets the current time (milliseconds) before queued acknowledgment messages are s
 static const char* set_ack_timeout_doc = R"docstring(set_ack_timeout(timeout)
 Sets the time before queued ack messages are sent.
 
+
 :param timeout: The maximum amount of time to wait after adding the first message to an acknowledgment
                 batch before sending the batch, in milliseconds. 0 indicates that the client will wait
                 until the batch is full. A value of 0 is not recommended unless the batch size is set to 1.
@@ -1151,6 +1207,7 @@ Enables or disables automatic retry of a command to AMPS after a reconnect. This
          a subscription manager, including all HAClients, will have all
          subscribe calls placed.
 
+
 :param enabled: False to disable automatic retry of commands to AMPS.
 :type enabled: Boolean
 )docstring";
@@ -1159,6 +1216,7 @@ Enables or disables automatic retry of a command to AMPS after a reconnect. This
 static const char* get_retry_on_disconnect_doc = R"docstring(get_retry_on_disconnect()
 Called to check if automatic retry of a command to AMPS after a reconnect is enabled.
 
+
 :returns: True, if automatic retry of a command to AMPS after a reconnect is enabled.
 )docstring";
 
@@ -1167,6 +1225,7 @@ static const char* set_default_max_depth_doc = R"docstring(set_default_max_depth
 Sets a default maximum depth for all new ``MessageStream`` objects
 that are returned from synchronous API calls such as :meth:`execute()`.
 
+
 :param depth: The new depth to use. A depth of ``0`` means no max and is the default.
 :type depth: int
 )docstring";
@@ -1175,6 +1234,7 @@ that are returned from synchronous API calls such as :meth:`execute()`.
 static const char* get_default_max_depth_doc = R"docstring(get_default_max_depth()
 Gets the maximum depth for any new ``MessageStream``.
 
+
 :returns: The current maximum depth for any new ``MessageStream``. ``0`` is no maximum.
 )docstring";
 
@@ -1183,6 +1243,7 @@ static const char* set_global_command_type_message_handler_doc = R"docstring(set
 Add a message handler to the :class:`Client` to handle messages from the
 server with the specified command.
 
+
 :param command: The command to send to the handler.
                 Valid values are ``ack`` and ``heartbeat``.
 :type command: str
@@ -1196,6 +1257,7 @@ static const char* get_error_on_publish_gap_doc = R"docstring(get_error_on_publi
 Indicates whether :exc:`PublishStoreGapException` can be thrown by the client publish store if the
 client logs onto a server that appears to be missing messages no longer held in the store.
 
+
 :returns: True if :exc:`PublishStoreGapException` can be thrown, False otherwise.
 
 )docstring";
@@ -1204,6 +1266,7 @@ client logs onto a server that appears to be missing messages no longer held in
 static const char* set_error_on_publish_gap_doc = R"docstring(set_error_on_publish_gap(error_on_publish_gap)
 Called to enable or disable throwing :exc:`PublishStoreGapException`.
 
+
 :param error_on_publish_gap: If True, :exc:`PublishStoreGapException` can be thrown by the
                              client publish store if the client logs onto a server that
                              appears to be missing messages no longer held in the store.
@@ -1213,6 +1276,7 @@ Called to enable or disable throwing :exc:`PublishStoreGapException`.
 static const char* add_http_preflight_header_doc = R"docstring(add_http_preflight_header(header)
 Called to add a header line to the HTTP preflight message sent via a proxy.
 
+
 :param header: The full header line to append to the HTTP upgrade message.
 :type header: string
 )docstring";
@@ -1220,6 +1284,7 @@ Called to add a header line to the HTTP preflight message sent via a proxy.
 static const char* add_http_preflight_header_key_value_doc = R"docstring(add_http_preflight_header_key_value(key, value)
 Called to add key: value as a header line to the HTTP preflight message sent via a proxy.
 
+
 :param key: The key in the header line to append to the HTTP upgrade message.
 :type key: string
 :param value: The value in the header line to append to the HTTP upgrade message.
@@ -1229,3 +1294,14 @@ Called to add key: value as a header line to the HTTP preflight message sent via
 static const char* clear_http_preflight_headers_doc = R"docstring(clear_http_preflight_headers()
 Called to clear all HTTP header lines previously set for preflight.
 )docstring";
+
+static const char* set_publish_batching_doc = R"docstring(set_publish_batching(batchSize, batchTimeoutMillis)
+Sets the max bytes to cache and max timeout in millis for caching delta_publish and publish commands.
+Most useful when your message data is small compared to tcp packet size.
+
+
+:param batchSize: The max number of bytes to cache before sending.
+:type batchSize: unsigned long
+:param batchTimeoutMillis: The max ms between sending batches.
+:type batchTimeoutMillis: unsigned long
+)docstring";
