@@ -717,7 +717,8 @@ namespace AMPS
   protected:
     static bool canResize(size_t requestedSize_, void* vpThis_)
     {
-      return ((BlockPublishStore*)vpThis_)->callResizeHandler(requestedSize_);
+      BlockPublishStore* me = (BlockPublishStore*)vpThis_;
+      return me->callResizeHandler(requestedSize_);
     }
 
     amps_uint64_t _getLowestUnpersisted() const
@@ -818,7 +819,7 @@ namespace AMPS
         pBuffer->setPosition(_metadataBlock->_offset);
         // Metadata block holds block size, block header size,
         // last discarded sequence, client version
-        pBuffer->putUint32((amps_uint32_t)blockSize);
+        pBuffer->putUint32(blockSize);
         pBuffer->putUint32((amps_uint32_t)getBlockHeaderSize());
         pBuffer->putUint64((amps_uint64_t)0);
         // Metadata blocks puts client version in CRC position
@@ -890,7 +891,7 @@ namespace AMPS
       if (blockHeader._blocksToWrite == 0)
       {
         pBuffer->setPosition(0);
-        pBuffer->putUint32((amps_uint32_t)blockSize);
+        pBuffer->putUint32(blockSize);
       }
       else
       {
@@ -1471,7 +1472,7 @@ namespace AMPS
       amps_uint32_t freeCount = 0;
       Block* firstFree = NULL;
       Block* endOfFreeList = NULL;
-      size_t blockSize = getBlockSize();
+      amps_uint32_t blockSize = getBlockSize();
       size_t numBlocks = size / blockSize;
       size_t blockNum = 0;
       // Used to create used list in order after recovery
