@@ -31,6 +31,7 @@
 #include <amps/Message.hpp>
 #include <amps/RecoveryPoint.hpp>
 #include <amps/RecoveryPointAdapter.hpp>
+#include <atomic>
 #include <functional>
 #include <map>
 #include <sstream>
@@ -1190,7 +1191,9 @@ namespace AMPS
       _recentChanged(true),
       _recovering(false),
       _recoveryPointAdapter(NULL),
-      _recoveryPointFactory(NULL)
+      _recoveryPointFactory(NULL),
+      _adapterSequence((amps_uint64_t)0),
+      _nextAdapterUpdate((amps_uint64_t)0)
     { ; }
 
     typedef RecoveryPointAdapter::iterator RecoveryIterator;
@@ -1211,6 +1214,8 @@ namespace AMPS
       , _recovering(true)
       , _recoveryPointAdapter(adapter_)
       , _recoveryPointFactory(factory_)
+      , _adapterSequence((amps_uint64_t)0)
+      , _nextAdapterUpdate((amps_uint64_t)0)
     {
       Message msg;
       if (!_recoveryPointFactory)
@@ -1626,8 +1631,25 @@ namespace AMPS
       }
       Field bookmark = pSub_->getMostRecentList(false);
       RecoveryPoint update = _recoveryPointFactory(pSub_->id(), bookmark);
+      // Use atomic seq to keep updates in order
+      amps_uint64_t seq = _adapterSequence.fetch_add(1);
       Unlock<Mutex> unlock(_lock);
-      _recoveryPointAdapter.update(update);
+      while (_nextAdapterUpdate.load() < seq)
+      {
+        AMPS_USLEEP(100);
+      }
+      try
+      {
+        _recoveryPointAdapter.update(update);
+      }
+      catch (const std::exception&)
+      {
+        _nextAdapterUpdate.fetch_add(1);
+        // Free memory, the bookmark needs to be cleared
+        bookmark.clear();
+        throw;
+      }
+      _nextAdapterUpdate.fetch_add(1);
       // Free memory, the bookmark needs to be cleared
       bookmark.clear();
     }
@@ -1640,6 +1662,8 @@ namespace AMPS
     typedef std::set<Subscription*> SubscriptionSet;
     RecoveryPointAdapter _recoveryPointAdapter;
     RecoveryPointFactory _recoveryPointFactory;
+    std::atomic<amps_uint64_t> _adapterSequence;
+    std::atomic<amps_uint64_t> _nextAdapterUpdate;
   };
 
 } // end namespace AMPS
