@@ -367,6 +367,20 @@ AMPSDLL amps_result amps_client_send_with_version(
   amps_handle message,
   unsigned* version_out);
 
+/**
+ *  Adds a message to the send cache, possibly sending the cache.
+ *  \param client The client to sent the message on.
+ *  \param message An AMPS message to send.
+ *  \param version_out The connection version used to send the message.
+ *  \param addToBatch Is 1 to add or 0 to send what's cached and this message.
+ *  \returns An amps_result indicating the success or failure of this send.
+ *    If not AMPS_E_OK, use amps_client_get_error() to find the text of the error message.
+ */
+AMPSDLL amps_result amps_client_send_batch(
+  amps_handle client,
+  amps_handle message,
+  unsigned* version_out,
+  int addToBatch);
 
 /**
  *  Sets the message handler function for this client
@@ -747,6 +761,16 @@ AMPSDLL amps_result amps_client_set_http_preflight_callback(
   amps_http_preflight_callback callback,
   void* userData);
 
+/**
+ * Sets a byte size batchSizeBytes and timeout for using batch sends of publish and
+ * delta_publish messages only.
+ * \param client_ The client on which to set the batchSizeBytess.
+ * \param batchSizeBytes_ The max number of bytes to cache before a send.
+ * \param batchTimeout_ The max time for cached messages to be held before send.
+ */
+AMPSDLL void amps_client_set_batch_send(amps_handle client_, amps_uint64_t batchSizeBytes_,
+                                        amps_uint64_t batchTimeout_);
+
 /**
  *
  * TRANSPORT-SPECIFIC APIS
@@ -800,6 +824,16 @@ AMPSDLL int amps_ssl_init(const char* dllPath_);
 
 AMPSDLL int amps_ssl_init_from_context(void* sslContext_, const char* fileName_);
 
+/* From OpenSSL header
+ * use either SSL_VERIFY_NONE or SSL_VERIFY_PEER, the last 3 options are
+ * 'ored' with SSL_VERIFY_PEER if they are desired
+# define SSL_VERIFY_NONE                 0x00
+# define SSL_VERIFY_PEER                 0x01
+# define SSL_VERIFY_FAIL_IF_NO_PEER_CERT 0x02
+# define SSL_VERIFY_CLIENT_ONCE          0x04
+# define SSL_VERIFY_POST_HANDSHAKE       0x08
+ */
+
 /**
  * Configures OpenSSL to validate the server's certificate when connecting.
  * This parameter must be set before attempting to connect to AMPS.
@@ -821,6 +855,36 @@ AMPSDLL int amps_ssl_set_verify(int mode_);
  */
 AMPSDLL int amps_ssl_load_verify_locations(const char* caFile_, const char* caPath_);
 
+/* From OpenSSL header
+# define SSL_FILETYPE_PEM       1
+# define SSL_FILETYPE_ASN1      2
+# define SSL_FILETYPE_DEFAULT   3
+*/
+
+/**
+ * Configures OpenSSL to use the specified file for its certificate.
+ *
+ * \param fileName_ A null-terminated string to pass to OpenSSL as the default PEM file.
+ * \param type_ An int representing the file type: SSL_FILETYPE_PEM, SSL_FILETYPE_ASN1, SSL_FILETYPE_DEFAULT
+ * \returns 0 if successful, -1 if an error occurred.
+ *
+ * See OpenSSL's SSL_CTX_use_certificate_file for more information on OpenSSL's certificate
+ * location options.
+ */
+AMPSDLL int amps_ssl_use_certificate_file(const char* fileName_, int type_);
+
+/**
+ * Configures OpenSSL to use the specified file for its private key.
+ *
+ * \param fileName_ A null-terminated string to pass to OpenSSL as the file.
+ * \param type_ An int representing the file type: AMPS_SSL_FILETYPE_PEM, AMPS_SSL_FILETYPE_ASN1, AMPS_SSL_FILETYPE_DEFAULT
+ * \returns 0 if successful, -1 if an error occurred.
+ *
+ * See OpenSSL's SSL_CTX_use_PrivateKey_file for more information on OpenSSL's certificate
+ * location options.
+ */
+AMPSDLL int amps_ssl_use_PrivateKey_file(const char* fileName_, int type_);
+
 /**
  * Frees OpenSSL context and shared library.
  * This function should be called to free up OpenSSL resources and context
