@@ -69,6 +69,8 @@ _amps_SSL_CTX*    (*_amps_SSL_CTX_new)(const _amps_SSL_METHOD*);
 void              (*_amps_SSL_CTX_free)(_amps_SSL_CTX*);
 void              (*_amps_SSL_CTX_set_verify)(_amps_SSL_CTX*, int, void*);
 int               (*_amps_SSL_CTX_load_verify_locations)(_amps_SSL_CTX*, const char*, const char*);
+int               (*_amps_SSL_CTX_use_certificate_file)(_amps_SSL_CTX*, const char*, int);
+int               (*_amps_SSL_CTX_use_PrivateKey_file)(_amps_SSL_CTX*, const char*, int);
 _amps_SSL*        (*_amps_SSL_new)(_amps_SSL_CTX*);
 int               (*_amps_SSL_set_fd)(_amps_SSL*, int);
 int               (*_amps_SSL_get_error)(_amps_SSL*, int);
@@ -257,6 +259,8 @@ int _amps_ssl_load(const char* dllPath_)
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_CTX_free);
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_CTX_set_verify);
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_CTX_load_verify_locations);
+  LOAD_AMPS_SSLEAY_FUNCTION(SSL_CTX_use_certificate_file);
+  LOAD_AMPS_SSLEAY_FUNCTION(SSL_CTX_use_PrivateKey_file);
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_new);
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_get_error);
   LOAD_AMPS_SSLEAY_FUNCTION(SSL_ctrl);
@@ -389,6 +393,8 @@ int _amps_ssl_load(const char* dllPath_)
   LOAD_AMPS_SSL_FUNCTION(SSL_CTX_free);
   LOAD_AMPS_SSL_FUNCTION(SSL_CTX_set_verify);
   LOAD_AMPS_SSL_FUNCTION(SSL_CTX_load_verify_locations);
+  LOAD_AMPS_SSL_FUNCTION(SSL_CTX_use_certificate_file);
+  LOAD_AMPS_SSL_FUNCTION(SSL_CTX_use_PrivateKey_file);
   LOAD_AMPS_SSL_FUNCTION(SSL_new);
   LOAD_AMPS_SSL_FUNCTION(SSL_ctrl);
   LOAD_AMPS_SSL_FUNCTION(SSL_get_error);
@@ -589,7 +595,41 @@ AMPSDLL int amps_ssl_load_verify_locations(const char* caFile_, const char* caPa
   }
   if (!_amps_SSL_CTX_load_verify_locations(_amps_ssl_ctx, caFile_, caPath_))
   {
-    _amps_ssl_set_error_from_stack("Unknown error setting CA certificate location.");
+    _amps_ssl_set_error_from_stack("Unknown error loading verify locations.");
+    return -1;
+  }
+  return 0;
+}
+
+AMPSDLL int amps_ssl_use_certificate_file(const char* fileName_, int type_)
+{
+  if (!_amps_ssl_ctx)
+  {
+    AMPS_strcpy(_amps_ssl_initialization_error, ERROR_BUFFER_LENGTH,
+                "amps_ssl_init must have been called successfully before setting this value.");
+    return -1;
+  }
+
+  if (!_amps_SSL_CTX_use_certificate_file(_amps_ssl_ctx, fileName_, type_))
+  {
+    _amps_ssl_set_error_from_stack("Unknown error setting certificate file.");
+    return -1;
+  }
+  return 0;
+}
+
+AMPSDLL int amps_ssl_use_PrivateKey_file(const char* fileName_, int type_)
+{
+  if (!_amps_ssl_ctx)
+  {
+    AMPS_strcpy(_amps_ssl_initialization_error, ERROR_BUFFER_LENGTH,
+                "amps_ssl_init must have been called successfully before setting this value.");
+    return -1;
+  }
+
+  if (!_amps_SSL_CTX_use_PrivateKey_file(_amps_ssl_ctx, fileName_, type_))
+  {
+    _amps_ssl_set_error_from_stack("Unknown error setting PrivateKey file.");
     return -1;
   }
   return 0;
