@@ -28,6 +28,7 @@
 #endif
 #include <amps/amps_impl.h>
 #include <amps/ampsuri.h>
+#include <amps/amps_zlib.h>
 #include <stdarg.h>
 #include <sys/types.h>
 
@@ -39,6 +40,7 @@
   #define AMPS_IEX_LONG(ptr, value) atomic_exchange_explicit((ptr), (value), memory_order_acq_rel)
   #define AMPS_FETCH_ADD(ptr, value) atomic_fetch_add_explicit((ptr), (value), memory_order_acq_rel)
   #define AMPS_FETCH_SUB(ptr, value) atomic_fetch_sub_explicit((ptr), (value), memory_order_acq_rel)
+  #define AMPS_ATOMIC_MODIFIER _Atomic
 #elif defined(_WIN32)
   #ifdef _WIN64
     #define AMPS_IEX(ptr,value) _InterlockedExchange64((LONG64*)(ptr), (LONG64)(value))
@@ -53,12 +55,14 @@
     #define AMPS_FETCH_ADD(ptr, value) InterlockedExchangeAdd((LONG volatile*)(ptr), (LONG)(value))
     #define AMPS_FETCH_SUB(ptr, value) InterlockedExchangeAdd((LONG volatile*)(ptr), (LONG)(-1 * (value)))
   #endif
+  #define AMPS_ATOMIC_MODIFIER volatile
 #else
   #define AMPS_IEX(ptr, value) (void)__sync_lock_test_and_set((ptr), (value))
   #define AMPS_IEX_GET(ptr, value) __sync_lock_test_and_set((ptr), (value))
   #define AMPS_IEX_LONG(ptr, value) (void)__sync_lock_test_and_set((ptr), (value))
   #define AMPS_FETCH_ADD(ptr, value) __sync_fetch_and_add((ptr), (value))
   #define AMPS_FETCH_SUB(ptr, value) __sync_fetch_and_sub((ptr), (value))
+  #define AMPS_ATOMIC_MODIFIER volatile
 #endif
 
 #ifdef _WIN32
@@ -392,11 +396,31 @@ typedef struct
   volatile pthread_t thread;
 #endif
 #endif
+
+  // Deflation (sending) support;
+#if __STDC_VERSION__ >= 201100
+  atomic_bool   useZlib;
+#else
+  volatile long useZlib;
+#endif
+  amps_zstream* pDeflater;
+  char*         pDeflateBuffer;
+  size_t        deflateBufferLength;
+  amps_zstream* pInflater;
+  char*         pInflateBuffer;
+  size_t        inflateBufferLength;
+  int           inflatePending;
+
   char lastErrorBuf[1024];
   amps_thread_exit_callback threadExitCallback;
   void* threadExitCallbackUserData;
   amps_http_preflight_callback httpPreflightCallback;
   void* httpPreflightCallbackUserData;
+  size_t sendBatchSize;
+  amps_uint64_t sendBatchTimeoutMillis;
+  char* sendCache;
+  size_t sendCacheOffset;
+  amps_uint64_t lastSendTime;
 }
 amps_tcp_t;
 
@@ -417,6 +441,13 @@ void amps_tcp_set_error(
   transport->lastErrorBuf[sizeof(transport->lastErrorBuf) - 1] = '\0';
 }
 
+void amps_tcp_set_zlib_error(amps_tcp_t* transport, int rc, const char* msg)
+{
+  _AMPS_SNPRINTF(transport->lastErrorBuf, sizeof(transport->lastErrorBuf),
+                 "zlib error code %d: %s", rc, transport->pDeflater->msg);
+  transport->lastErrorBuf[sizeof(transport->lastErrorBuf) - 1] = '\0';
+}
+
 void amps_tcp_set_socket_error(
   amps_tcp_t* transport)
 {
@@ -464,6 +495,105 @@ void amps_tcp_atfork_handler(amps_tcp_t* transport_, int code_)
   }
 }
 
+int amps_tcp_send_compressed(amps_tcp_t* me, amps_char* data, size_t length)
+{
+  amps_zstream* pDeflater = me->pDeflater;
+
+  pDeflater->next_in = data;
+  pDeflater->avail_in = (unsigned int)length;
+  pDeflater->avail_out = (unsigned int)me->deflateBufferLength;
+
+  // The ... || !pDeflater->avail_out is so we can send the flushed bytes
+  // even if all the input data is consumed, when the output buffer fills up.
+  while (pDeflater->avail_in || !pDeflater->avail_out)
+  {
+    // We send the whole deflate buffer after every deflate().
+    pDeflater->avail_out = (unsigned int)me->deflateBufferLength;
+    pDeflater->next_out = me->pDeflateBuffer;
+
+    int rc = amps_deflate(pDeflater, 2);
+    if (rc != 0)
+    {
+      amps_tcp_set_zlib_error(me, rc, pDeflater->msg);
+      return -1;
+    }
+
+    // Send the whole deflate buffer here.
+    const char* p = me->pDeflateBuffer, *pe = pDeflater->next_out;
+    while (p != pe)
+    {
+#ifndef _WIN32
+      ssize_t sentBytes = send(me->fd, p, (size_t)(pe - p), MSG_NOSIGNAL);
+#else
+      int sentBytes = send(me->fd, p, (int)(pe - p), 0);
+#endif
+      if (sentBytes <= 0)
+      {
+        return -1;
+      }
+      p += sentBytes;
+    }
+  }
+  return (int)length;
+}
+
+int amps_tcp_recv_compressed(amps_tcp_t* me, size_t capacity)
+{
+  amps_zstream* pInflater = me->pInflater;
+  int bytes = 0;
+
+  if (!pInflater->avail_in && !me->inflatePending)
+  {
+    errno = 0;
+    // Retrieve data from the socket only if there's nothing to do.
+#ifndef _WIN32
+    bytes = (int)recv(me->fd, me->pInflateBuffer, me->inflateBufferLength, 0);
+#else
+    bytes = (int)recv(me->fd, me->pInflateBuffer, (int)me->inflateBufferLength, 0);
+#endif
+    if (bytes <= 0)
+    {
+      return bytes;
+    }
+    pInflater->avail_in = (unsigned int)bytes;
+    pInflater->next_in = me->pInflateBuffer;
+  }
+  pInflater->avail_out = (unsigned int)capacity;
+
+  if (pInflater->avail_out)
+  {
+    me->inflatePending = 0;
+    int rc = amps_inflate(pInflater, 2);
+    if (rc != 0)
+    {
+      if (rc > 0)
+      {
+        return -1;
+      }
+      return rc;
+    }
+    // If we got no output, we need to read more input
+    if (pInflater->avail_out == capacity)
+    {
+      return AMPS_ZLIB_WANT_READ;
+    }
+    if (pInflater->avail_out == 0)
+    {
+      // Special case where, even though we consumed the input, we have more output to write next time someone calls here.
+      me->inflatePending = 1;
+      // Are we out of space to inflate?
+      if (pInflater->avail_in
+          && pInflater->next_in < (me->pInflateBuffer + bytes))
+      {
+        return AMPS_ZLIB_NEED_SPACE;
+      }
+    }
+    return (int)(capacity - pInflater->avail_out);
+  }
+
+  return 0;
+}
+
 amps_handle amps_tcp_create(void)
 {
   amps_tcp_t* transport = (amps_tcp_t*)malloc(sizeof(amps_tcp_t));
@@ -478,6 +608,9 @@ amps_handle amps_tcp_create(void)
 #if __STDC_VERSION__ >= 201100
     transport->destroying = false;
     transport->disconnecting = false;
+#else
+    AMPS_IEX_LONG(&transport->destroying, 0);
+    AMPS_IEX_LONG(&transport->disconnecting, 0);
 #endif
     amps_atfork_add(transport, (_amps_atfork_callback_function)amps_tcp_atfork_handler);
   }
@@ -490,6 +623,68 @@ amps_handle amps_tcp_create(void)
   void* amps_tcp_threaded_reader(void* userData);
 #endif
 
+amps_result
+amps_tcp_init_zlib(amps_tcp_t* me)
+{
+  const int windowBits = -15; /* no headers */
+  if (me->pDeflater)
+  {
+    amps_deflateEnd(me->pDeflater);
+    free(me->pDeflater);
+    me->pDeflater = 0;
+
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
+  }
+  me->pDeflater = (amps_zstream*)malloc(sizeof(amps_zstream));
+  memset(me->pDeflater, 0, sizeof(amps_zstream));
+  int rc = amps_deflateInit2(me->pDeflater,
+                  6,   /* level */
+                  8,   /* Z_DEFLATED */
+                  windowBits,
+                  8,   /* default memory level */
+                  0);  /* default strategy */
+  if (rc != 0)
+  {
+    free(me->pDeflater);
+    me->pDeflater = 0;
+    amps_tcp_set_error(me, "Unable to initialize zlib.");
+    return AMPS_E_URI;
+  }
+  me->pInflater = (amps_zstream*)malloc(sizeof(amps_zstream));
+  memset(me->pInflater, 0, sizeof(amps_zstream));
+  rc = amps_inflateInit2(me->pInflater, windowBits)
+  if (rc != 0)
+  {
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
+    amps_tcp_set_error(me, "Unable to initialize zlib.");
+    return AMPS_E_URI;
+  }
+
+  //  Initialize the work buffers for inflation and deflation;
+  if (!me->pDeflateBuffer)
+  {
+    me->pDeflateBuffer = (char*)malloc(16384);
+    me->deflateBufferLength = 16384;
+  }
+
+  if (!me->pInflateBuffer)
+  {
+    me->pInflateBuffer = (char*)malloc(16384);
+    me->inflateBufferLength = 16384;
+  }
+  me->inflatePending = 0;
+#if __STDC_VERSION__ >= 201100
+  me->useZlib = true;
+#else
+  AMPS_IEX_LONG(&me->useZlib, 1);
+#endif
+  return AMPS_E_OK;
+}
+
 int
 amps_tcp_opt_parse(const char* val, size_t valLen, int* parsed)
 {
@@ -525,9 +720,10 @@ amps_tcp_opt_parse(const char* val, size_t valLen, int* parsed)
 
 typedef enum
 {
-  AMPS_PROPERTY_OK,
-  AMPS_PROPERTY_PREFLIGHT_TRUE,
-  AMPS_PROPERTY_PARSE_ERROR
+  AMPS_PROPERTY_PARSE_ERROR = 0,
+  AMPS_PROPERTY_OK = 1,
+  AMPS_PROPERTY_PREFLIGHT_TRUE = 2,
+  AMPS_PROPERTY_COMPRESSION_ZLIB = 4,
 } amps_property_result;
 
 amps_property_result
@@ -642,6 +838,12 @@ amps_tcp_apply_socket_property(AMPS_SOCKET fd, const char* key, size_t keyLen, c
     // no-op; handled in the connect function
     return AMPS_PROPERTY_OK;
   }
+  else if (keyLen == 11 && memcmp(key, "compression", keyLen) == 0)
+  {
+    return (valLen == 4 && memcmp(val, "zlib", 4) == 0)
+            ? AMPS_PROPERTY_COMPRESSION_ZLIB
+            : AMPS_PROPERTY_PARSE_ERROR;
+  }
   // properties after this point must only be boolean or numeric
   if (!amps_tcp_opt_parse(val, valLen, &value))
   {
@@ -698,7 +900,7 @@ amps_property_result
 amps_tcp_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
                                  size_t uriLength, amps_uri_state* uriState)
 {
-  amps_property_result retVal = AMPS_PROPERTY_OK;
+  int retVal = AMPS_PROPERTY_OK;
   amps_property_result optRet = AMPS_PROPERTY_OK;
   const char* key = NULL;
   size_t keyLength = 0;
@@ -730,7 +932,7 @@ amps_tcp_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
         {
           return optRet;
         }
-        retVal = optRet;
+        retVal |= (int)optRet;
       }
     }
   }
@@ -738,7 +940,7 @@ amps_tcp_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
   {
     return AMPS_PROPERTY_PARSE_ERROR;
   }
-  return retVal;
+  return (amps_property_result)retVal;
 }
 
 amps_result _amps_tcp_handle_http_preflight(amps_tcp_t* me)
@@ -792,6 +994,7 @@ amps_result _amps_tcp_handle_http_preflight(amps_tcp_t* me)
   socketBytes = 0;
   while ((readPoint - me->buf) < 15)
   {
+    errno = 0;
 #ifdef _WIN32
     socketBytes = recv(me->fd, (char*)readPoint, (int)(end - readPoint), 0);
 #else
@@ -832,6 +1035,7 @@ amps_result _amps_tcp_handle_http_preflight(amps_tcp_t* me)
   readPoint = me->buf;
   while (socketBytes > 0)
   {
+    errno = 0;
 #ifdef _WIN32
     socketBytes = recv(me->fd, (char*)readPoint, (int)AMPS_TCP_DEFAULT_BUF_LEN, 0);
 #else
@@ -881,7 +1085,7 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
   amps_result result = AMPS_E_OK;
   struct addrinfo* pAddrInfo = NULL, addrhints;
   size_t address_length = 0;
-  AMPS_SOCKET fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
+  AMPS_ATOMIC_MODIFIER AMPS_SOCKET fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
   amps_property_result propResult;
   int i;
 
@@ -898,6 +1102,7 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me->lock);
 #endif
+  me->lastErrorBuf[0] = '\0';
   /* if we were previously marked as disconnecting, forget that. */
 #if __STDC_VERSION__ >= 201100
   me->disconnecting = false;
@@ -1014,8 +1219,8 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
     goto error;
   }
 
-  me->fd = socket(pAddrInfo->ai_family, pAddrInfo->ai_socktype, pAddrInfo->ai_protocol);
-  if (me->fd == AMPS_INVALID_SOCKET)
+  fd = socket(pAddrInfo->ai_family, pAddrInfo->ai_socktype, pAddrInfo->ai_protocol);
+  if (fd == AMPS_INVALID_SOCKET)
   {
     freeaddrinfo(pAddrInfo);
     amps_tcp_set_socket_error(me);
@@ -1025,33 +1230,67 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
 
 
   /* apply socket properties */
-  propResult = amps_tcp_apply_socket_properties(me->fd, address, address_length,
+  propResult = amps_tcp_apply_socket_properties(fd, address, address_length,
                                                 &uri_state_socket_props);
+#if __STDC_VERSION__ >= 201100
+  me->useZlib = false;
+#else
+  AMPS_IEX_LONG(&me->useZlib, 0);
+#endif
   if (propResult != AMPS_PROPERTY_OK)
   {
     if (propResult == AMPS_PROPERTY_PARSE_ERROR)
     {
       freeaddrinfo(pAddrInfo);
-      fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
-      amps_tcp_set_error(me, "The URI specified invalid properties.");
+      if (me->lastErrorBuf[0] == '\0')
+      {
+        char errBuf[128];
+        _AMPS_SNPRINTF(errBuf, 128, "The URI specified invalid properties: %.*s.", (int)uri_state_socket_props.part_length, uri_state_socket_props.part);
+        amps_tcp_set_error(me, errBuf);
+      }
       result = AMPS_E_URI;
       goto error;
     }
+    if ((propResult & AMPS_PROPERTY_COMPRESSION_ZLIB))
+    {
+      if (amps_zlib_init(NULL) == -1)
+      {
+        amps_tcp_set_error(me, amps_zlib_last_error);
+        freeaddrinfo(pAddrInfo);
+        result = AMPS_E_URI;
+#if __STDC_VERSION__ >= 201100
+        me->useZlib = true;
+#else
+        AMPS_IEX_LONG(&me->useZlib, 1);
+#endif
+        goto error;
+      }
+      if (amps_tcp_init_zlib(me) != AMPS_E_OK)
+      {
+        freeaddrinfo(pAddrInfo);
+        result = AMPS_E_URI;
+#if __STDC_VERSION__ >= 201100
+        me->useZlib = true;
+#else
+        AMPS_IEX_LONG(&me->useZlib, 1);
+#endif
+        goto error;
+      }
+    }
   }
 #ifdef __APPLE__
   rc = 1; /* Set NOSIGPIPE to ON */
-  if (setsockopt(me->fd, SOL_SOCKET, SO_NOSIGPIPE, &rc, sizeof(rc)) < 0)
+  if (setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &rc, sizeof(rc)) < 0)
   {
     perror ("setsockopt(,,SO_NOSIGPIPE)");
     freeaddrinfo(pAddrInfo);
-    fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
     amps_tcp_set_error(me, "Failed to set no SIGPIPE.");
     result = AMPS_E_CONNECTION;
     goto error;
   }
 #endif
 
-  rc = connect(me->fd, pAddrInfo->ai_addr,
+  rc = connect(fd, pAddrInfo->ai_addr,
                (socklen_t)pAddrInfo->ai_addrlen);
   freeaddrinfo(pAddrInfo);
 #ifdef _WIN32
@@ -1060,11 +1299,12 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
   if (rc == -1)
 #endif
   {
-    fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
     amps_tcp_set_socket_error(me);
     result = AMPS_E_CONNECTION_REFUSED;
     goto error;
   }
+  /* Now set fd on me */
+  me->fd = AMPS_IEX_GET(&fd, AMPS_INVALID_SOCKET);
 
   /* Increase the connection version now, so that anyone else attempting a
    * reconnect knows that they may not need to. */
@@ -1072,11 +1312,12 @@ amps_result amps_tcp_connect(amps_handle transport, const amps_char* address)
   result = AMPS_E_OK;
 
   // Handle http_preflight upgrade from HTTP
-  if (propResult == AMPS_PROPERTY_PREFLIGHT_TRUE && me->httpPreflightCallback)
+  if ((propResult & AMPS_PROPERTY_PREFLIGHT_TRUE) && me->httpPreflightCallback)
   {
     result = _amps_tcp_handle_http_preflight(me);
     if (result != AMPS_E_OK)
     {
+      fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
       if (fd != AMPS_INVALID_SOCKET && !me->disconnecting)
       {
         shutdown(fd, SHUT_RDWR);
@@ -1249,6 +1490,25 @@ void amps_tcp_destroy(amps_handle transport)
   AMPS_JOIN(me->thread);
   AMPS_SLEEP(1);
   free(me->buf);
+  if(me->pDeflater)
+  {
+    amps_deflateEnd(me->pDeflater);
+    free(me->pDeflater);
+    me->pDeflater = 0;
+
+    free(me->pDeflateBuffer);
+    me->pDeflateBuffer = 0;
+  }
+  if(me->pInflater)
+  {
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
+
+    free(me->pInflateBuffer);
+    me->pInflateBuffer = 0;
+  }
+  free(me->sendCache);
   /* Hopefully, nobody else is using me right now. */
   AMPS_KILLLOCK(&me->lock);
   AMPS_KILLLOCK(&me->sendLock);
@@ -1343,11 +1603,57 @@ void amps_tcp_handle_stream_corruption(
   amps_tcp_handle_disconnect(me, failedConnectionVersion);
 }
 
+int amps_tcp_send_bytes(amps_tcp_t* me, char* buf, int bytesWritten)
+{
+  int bytesSent = 0;
+  while (bytesSent < bytesWritten)
+  {
+    int bytesWrittenThisTime;
+    if (me->fd == AMPS_INVALID_SOCKET)
+    {
+      amps_tcp_set_error(me, "Not connected.");
+      AMPS_UNLOCK(&me->sendLock);
+      return AMPS_E_DISCONNECTED;
+    }
+
+    if (me->useZlib)
+    {
+      bytesWrittenThisTime = amps_tcp_send_compressed(me, (buf)+bytesSent, (size_t)(bytesWritten - bytesSent));
+    }
+    else
+    {
+#ifdef _WIN32
+      bytesWrittenThisTime = send(me->fd, buf + bytesSent, bytesWritten - bytesSent, 0);
+#else
+      bytesWrittenThisTime = (int)send(me->fd, buf + bytesSent, (size_t)(bytesWritten - bytesSent), MSG_NOSIGNAL);
+#endif
+    }
+    if (bytesWrittenThisTime <= 0)
+    {
+      return -1;
+    }
+    bytesSent += bytesWrittenThisTime;
+  }
+  return bytesSent;
+}
+
+int
+amps_tcp_send_cache(amps_tcp_t* me)
+{
+  me->lastSendTime = amps_now();
+  int bytesSent = amps_tcp_send_bytes(me, me->sendCache, (int)(me->sendCacheOffset));
+  if (bytesSent >= 0)
+  {
+      me->sendCacheOffset = 0;
+  }
+  return bytesSent;
+}
 
 amps_result
-amps_tcp_send_with_version(amps_handle transport,
-                           amps_handle message,
-                           unsigned* version_out)
+amps_tcp_send_batch(amps_handle transport,
+                    amps_handle message,
+                    unsigned* version_out,
+                    int addToBatch)
 {
   amps_tcp_t* me = (amps_tcp_t*)transport;
   size_t len = AMPS_TCP_DEFAULT_BUF_LEN;
@@ -1373,65 +1679,149 @@ amps_tcp_send_with_version(amps_handle transport,
   AMPS_LOCK(&me->sendLock);
   while (bytesWritten < 0)
   {
-    if (me->buf == NULL)
+    if (me->sendCache && addToBatch == 0)
     {
-      me->buf = (char*)malloc(len);
-      if (me->buf == NULL)
+      // Send the cache
+      bytesSent = amps_tcp_send_cache(me);
+      if (bytesSent < 0)
       {
-        amps_tcp_set_error(me, "Unable to allocate memory to send message.");
+          /* record the error */
+          amps_tcp_set_error(me, "The connection is closed.");
+          AMPS_UNLOCK(&me->sendLock);
+          return AMPS_E_DISCONNECTED;
+      }
+    }
+    if (addToBatch && me->sendBatchSize)
+    {
+      // Serialize into cache
+      if (me->sendCache == NULL)
+      {
+        me->sendCache = (char*)malloc(me->sendBatchSize);
+        me->sendCacheOffset = 0;
+        if (me->sendCache == NULL)
+        {
+          amps_tcp_set_error(me, "Unable to allocate memory to cache message.");
+          AMPS_UNLOCK(&me->sendLock);
+          return AMPS_E_MEMORY;
+        }
+      }
+      bytesWritten = amps_message_serialize(message, me->serializer,
+                                            (me->sendCache) + me->sendCacheOffset + 4,
+                                            me->sendBatchSize - me->sendCacheOffset - 4);
+      /* amps_message_serialize could have written 0 bytes.
+       * who are we to judge?  If it returns less than 0,
+       * there wasn't enough room. */
+      if (bytesWritten >= 0)
+      {
+        // record the message length all up in the first 4 bytes.
+        bytesWrittenN = htonl((unsigned int)bytesWritten);
+        me->filterFunction((const unsigned char*)(me->sendCache + me->sendCacheOffset + 4),
+                           (size_t)bytesWritten, 0, me->filterUserData);
+        *((unsigned int*)(me->sendCache + me->sendCacheOffset)) = bytesWrittenN;
+        bytesWritten += 4;
+        if (!me->sendCacheOffset)
+        {
+          // Start the send clock based on first cached message
+          me->lastSendTime = amps_now();
+        }
+        me->sendCacheOffset += (size_t)bytesWritten;
+        if (me->sendCacheOffset < (me->sendBatchSize - 32)
+            && me->lastSendTime > (amps_now() - me->sendBatchTimeoutMillis))
+        {
+          AMPS_UNLOCK(&me->sendLock);
+          return AMPS_E_OK;
+        }
+        bytesSent = amps_tcp_send_cache(me);
+        if (bytesSent < 0)
+        {
+            /* record the error */
+            amps_tcp_set_error(me, "The connection is closed.");
+            AMPS_UNLOCK(&me->sendLock);
+            return AMPS_E_DISCONNECTED;
+        }
         AMPS_UNLOCK(&me->sendLock);
-        return AMPS_E_MEMORY;
+        return AMPS_E_OK;
+      }
+      // Need more space
+      if (me->sendCacheOffset)
+      {
+        // Send cache and try again
+        bytesSent = amps_tcp_send_cache(me);
+        if (bytesSent < 0)
+        {
+            /* record the error */
+            amps_tcp_set_error(me, "The connection is closed.");
+            AMPS_UNLOCK(&me->sendLock);
+            return AMPS_E_DISCONNECTED;
+        }
+      }
+      else
+      {
+        // Message is larger than batch max bytes, so just send it
+        addToBatch = 0;
       }
-      me->capacity = len;
     }
     else
     {
-      len = me->capacity;
+      if (me->buf == NULL)
+      {
+        me->buf = (char*)malloc(len);
+        if (me->buf == NULL)
+        {
+          amps_tcp_set_error(me, "Unable to allocate memory to send message.");
+          AMPS_UNLOCK(&me->sendLock);
+          return AMPS_E_MEMORY;
+        }
+        me->capacity = len;
+      }
+      else
+      {
+        len = me->capacity;
+      }
+      /* reserve 4 bytes for the length of the message */
+      bytesWritten = amps_message_serialize(message, me->serializer, (me->buf) + 4, len - 4);
+      /* amps_message_serialize could have written 0 bytes.
+       * who are we to judge?  If it returns less than 0,
+       * there wasn't enough room. */
+      if (bytesWritten >= 0)
+      {
+        // record the message length all up in the first 4 bytes.
+        bytesWrittenN = htonl((unsigned int)bytesWritten);
+        me->filterFunction((const unsigned char*)(me->buf + 4), (size_t)bytesWritten, 0, me->filterUserData);
+        *((unsigned int*)(me->buf)) = bytesWrittenN;
+        bytesWritten += 4;
+        break;
+      }
+      /* free this buffer, allocate a larger buffer next time */
+      free(me->buf);
+      me->capacity = 0;
+      me->buf = NULL;
+      len = (size_t)((double)len * 1.5);
     }
-    /* reserve 4 bytes for the length of the message */
-    bytesWritten = amps_message_serialize(message, me->serializer, (me->buf) + 4, len - 4);
-    /* amps_message_serialize could have written 0 bytes.
-     * who are we to judge?  If it returns less than 0,
-     * there wasn't enough room. */
-    if (bytesWritten >= 0)
-    {
-      break;
-    }
-    /* free this buffer, allocate a larger buffer next time */
-    free(me->buf);
-    me->capacity = 0;
-    me->buf = NULL;
-    len = (size_t)((double)len * 1.5);
   }
-  /* once we're done, we don't free buf -- it hangs around until we need it next time.
-   * record the message length all up in the first 4 bytes. */
-  bytesWrittenN = htonl((unsigned int)bytesWritten);
-  me->filterFunction((const unsigned char*)(me->buf + 4), (size_t)bytesWritten, 0, me->filterUserData);
-  *((unsigned int*)(me->buf)) = bytesWrittenN;
-  bytesWritten += 4;
+  /* once we're done, we don't free buf -- it hangs around until we need it next time. */
   /*now, send */
-  while (bytesSent < bytesWritten)
+  bytesSent = amps_tcp_send_bytes(me, me->buf, bytesWritten);
+  if (bytesSent < 0)
   {
-    int bytesWrittenThisTime;
-#ifdef _WIN32
-    bytesWrittenThisTime = send(me->fd, (me->buf) + bytesSent, bytesWritten - bytesSent, 0);
-#else
-    bytesWrittenThisTime = (int)send(me->fd, (me->buf) + bytesSent, (size_t)(bytesWritten - bytesSent), MSG_NOSIGNAL);
-#endif
-    if (bytesWrittenThisTime <= 0)
-    {
       /* record the error */
       amps_tcp_set_error(me, "The connection is closed.");
       AMPS_UNLOCK(&me->sendLock);
       return AMPS_E_DISCONNECTED;
-    }
-    bytesSent += bytesWrittenThisTime;
   }
   AMPS_UNLOCK(&me->sendLock);
 
   return AMPS_E_OK;
 }
 
+amps_result
+amps_tcp_send_with_version(amps_handle transport,
+                           amps_handle message,
+                           unsigned* version_out)
+{
+  return amps_tcp_send_batch(transport, message, version_out, 0);
+}
+
 amps_result
 amps_tcp_send(amps_handle transport,
               amps_handle message)
@@ -1500,6 +1890,7 @@ void* amps_tcp_threaded_reader(void* userData)
   int cancelState = 0;
   const size_t BUFFER_SIZE = AMPS_TCP_DEFAULT_BUF_LEN;
   const unsigned MAX_MESSAGE_SIZE = 1024 * 1024 * 1024;
+  amps_uint64_t now = 0;
   amps_uint64_t lastReadTime = 0;
   amps_uint64_t lastIdleTime = 0;
   amps_handle   message;
@@ -1542,10 +1933,11 @@ void* amps_tcp_threaded_reader(void* userData)
     {
       goto cleanup;
     }
+    now = amps_now();
     if (me->idleTimeMillis > 0 &&
-        (amps_now() - lastIdleTime) > me->idleTimeMillis)
+        (now - lastIdleTime) > me->idleTimeMillis)
     {
-      lastIdleTime = amps_now();
+      lastIdleTime = now;
 #ifndef _WIN32
       cancelState = 0;
       pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
@@ -1556,14 +1948,100 @@ void* amps_tcp_threaded_reader(void* userData)
       pthread_setcancelstate(cancelState, &unusedCancelState);
 #endif
     }
-#ifdef _WIN32
-    received = recv(me->fd, (char*)readPoint, (int)(end - readPoint), 0);
-#else
-    received = recv(me->fd, (char*)readPoint, (size_t)(end - readPoint), 0);
+    if (me->sendBatchTimeoutMillis
+        && (now - me->lastSendTime) > me->sendBatchTimeoutMillis
+        && me->sendCacheOffset)
+    {
+      AMPS_LOCK(&me->sendLock);
+      int bytesSent = amps_tcp_send_cache(me);
+      AMPS_UNLOCK(&me->sendLock);
+      if (bytesSent < 0)
+      {
+        if (fd != AMPS_INVALID_SOCKET && !me->disconnecting)
+        {
+          shutdown(fd, SHUT_RDWR);
+          amps_tcp_handle_disconnect(me, connectionVersion);
+        }
+        goto cleanup;
+      }
+    }
+    if (me->useZlib)
+    {
+      // Make sure that next_out is set
+      if (!me->disconnecting
+          && !me->destroying
+          && !me->pInflater->next_out)
+      {
+        me->pInflater->next_out = (const char*)readPoint;
+      }
+      received = amps_tcp_recv_compressed(me, (size_t)((const char*)end-me->pInflater->next_out));
+
+      if (received == AMPS_ZLIB_WANT_READ)
+      {
+        // Need to loop again before we try to parse anything
+        continue;
+      }
+      else if (received == AMPS_ZLIB_NEED_SPACE)
+      {
+        // Add space to the buffer for inflated messages
+        size_t newLength = 2 * (size_t)(end - buffer);
+        newBuffer = (unsigned char*)malloc( newLength );
+        if (newBuffer == NULL)
+        {
+          /* stream broken */
+          shutdown(fd, SHUT_RDWR);
+          amps_tcp_handle_disconnect(me, connectionVersion);
+          goto cleanup;
+        }
+        memcpy(newBuffer, buffer, (size_t)(end - buffer));
+        readPoint = newBuffer + (size_t)((unsigned const char*)readPoint - buffer);
+        if (!me->disconnecting
+            && !me->destroying)
+        {
+          me->pInflater->next_out = (const char*)newBuffer
+                                    + (size_t)((unsigned const char*)me->pInflater->next_out - buffer);
+        }
+        parsePoint = newBuffer + (size_t)(parsePoint - buffer);
+#ifndef _WIN32
+        cancelState = 0;
+        pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
 #endif
+        free(buffer);
+        buffer = newBuffer;
+#ifndef _WIN32
+        pthread_setcancelstate(cancelState, &unusedCancelState);
+#endif
+        end = newBuffer + newLength;
+        assert(readPoint >= parsePoint);
+        continue;
+      }
+    }
+    else
+    {
+      errno = 0;
+#ifdef _WIN32
+      received = recv(me->fd, (char*)readPoint, (int)(end - readPoint), 0);
+#else
+      received = recv(me->fd, (char*)readPoint, (size_t)(end - readPoint), 0);
+#endif
+    }
     if (received > 0)
     {
       /* Call filter function without the 4 byte size */
+      if (me->useZlib
+          && !me->disconnecting
+          && !me->destroying)
+      {
+        // Because we could have looped on recv_compressed and gotten smaller
+        // values each time, get the full received size.
+#ifdef _WIN32
+        received = (int)(me->pInflater->next_out - (const char*)readPoint);
+#else
+        received = me->pInflater->next_out - (const char*)readPoint;
+#endif
+        // We can clear next_out as we're done with current inflate
+        me->pInflater->next_out = NULL;
+      }
       me->filterFunction(readPoint, (size_t)received, 1, me->filterUserData);
       readPoint += received;
       lastReadTime = amps_now();
@@ -1673,6 +2151,14 @@ void* amps_tcp_threaded_reader(void* userData)
         memcpy(newBuffer, parsePoint, (size_t)(readPoint - parsePoint));
         readPoint = newBuffer + (readPoint - parsePoint);
         parsePoint = newBuffer;
+        // Keep next_out valid as well
+        if (!me->disconnecting
+            && !me->destroying
+            && me->pInflater
+            && me->pInflater->next_out)
+        {
+          me->pInflater->next_out = (const char*)newBuffer + (size_t)((unsigned const char*)me->pInflater->next_out - buffer);
+        }
 #ifndef _WIN32
         cancelState = 0;
         pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
@@ -1693,7 +2179,9 @@ void* amps_tcp_threaded_reader(void* userData)
         break;
       }
     } /*while(readPoint >= parsePoint + 4) */
-    if (readPoint > parsePoint)
+    // Create more reading room if we have left-over bytes and have consumed the beginning bytes
+    if (readPoint > parsePoint
+        && parsePoint > buffer)
     {
       memmove(buffer, parsePoint, (size_t)(readPoint - parsePoint));
     }
@@ -1782,12 +2270,15 @@ amps_tcp_get_socket(amps_handle transport)
 amps_result
 amps_tcp_update_read_timeout(amps_tcp_t* me)
 {
-  int timeout = (int)((me->readTimeoutMillis && me->idleTimeMillis) ?
-                      AMPS_MIN(me->readTimeoutMillis, me->idleTimeMillis) :
-                      AMPS_MAX(me->readTimeoutMillis, me->idleTimeMillis));
+  amps_uint64_t timeout = (me->readTimeoutMillis && me->idleTimeMillis)
+                          ? AMPS_MIN(me->readTimeoutMillis, me->idleTimeMillis)
+                          : AMPS_MAX(me->readTimeoutMillis, me->idleTimeMillis);
+  timeout = (timeout && me->sendBatchTimeoutMillis)
+            ? AMPS_MIN(timeout, me->sendBatchTimeoutMillis)
+            : AMPS_MAX(timeout, me->sendBatchTimeoutMillis);
   int rc = 0;
 #ifdef _WIN32
-  DWORD timeoutMillis = timeout;
+  DWORD timeoutMillis = (DWORD)timeout;
   if (me->fd == AMPS_INVALID_SOCKET)
   {
     return AMPS_E_DISCONNECTED;
@@ -1796,8 +2287,8 @@ amps_tcp_update_read_timeout(amps_tcp_t* me)
                   (const char*)&timeoutMillis, sizeof(DWORD));
 #else
   struct timeval tv;
-  tv.tv_sec = timeout / 1000;
-  tv.tv_usec = (timeout % 1000) * 1000;
+  tv.tv_sec = (int)timeout / 1000;
+  tv.tv_usec = ((int)timeout % 1000) * 1000;
   if (me->fd == AMPS_INVALID_SOCKET)
   {
     return AMPS_E_DISCONNECTED;
@@ -1864,3 +2355,13 @@ amps_tcp_set_idle_time(amps_handle transport, int millis)
   me->idleTimeMillis = (amps_uint64_t)millis;
   return amps_tcp_update_read_timeout(me);
 }
+
+void
+amps_tcp_set_batch_send(amps_handle transport, amps_uint64_t batchSize, amps_uint64_t timeout)
+{
+  amps_tcp_t* me = (amps_tcp_t*)transport;
+  me->sendBatchSize = batchSize;
+  me->sendBatchTimeoutMillis = timeout;
+  amps_tcp_update_read_timeout(me);
+}
+
