@@ -29,6 +29,7 @@
 #include <amps/amps_impl.h>
 #include <amps/amps_ssl.h>
 #include <amps/ampsuri.h>
+#include <amps/amps_zlib.h>
 #if __STDC_VERSION__ >= 201100
   #include <stdatomic.h>
   #include <stdbool.h>
@@ -303,70 +304,38 @@ typedef struct
   volatile pthread_t thread;
 #endif
 #endif
+
+  // Deflation (sending) support;
+#if __STDC_VERSION__ >= 201100
+  atomic_bool   useZlib;
+#else
+  volatile long useZlib;
+#endif
+  amps_zstream* pDeflater;
+  char*         pDeflateBuffer;
+  size_t        deflateBufferLength;
+  amps_zstream* pInflater;
+  char*         pInflateBuffer;
+  size_t        inflateBufferLength;
+  int           inflatePending;
+
   char lastErrorBuf[1024];
   amps_thread_exit_callback threadExitCallback;
   void* threadExitCallbackUserData;
   amps_http_preflight_callback httpPreflightCallback;
   void* httpPreflightCallbackUserData;
+  amps_uint64_t sendBatchSizeBytes;
+  amps_uint64_t sendBatchTimeoutMillis;
+  char* sendCache;
+  size_t sendCacheOffset;
+  amps_uint64_t lastSendTime;
 }
 amps_tcps_t;
 
-// Function to close SSL* and fd that can be replaced by Python client
-int _amps_tcps_close_all(AMPS_SOCKET fd, _amps_SSL* ssl)
+void amps_tcps_set_zlib_error(amps_tcps_t* transport, int rc)
 {
-  int rc = 0;
-  if (ssl)
-  {
-    rc = _amps_SSL_shutdown(ssl);
-    if (rc == 0 && fd != AMPS_INVALID_SOCKET)
-    {
-      // attempt to wait for the server to send back the matching close notify.
-#ifdef _WIN32
-      struct timeval tv;
-      fd_set readfds, exceptfds;
-      tv.tv_sec = 0;
-      tv.tv_usec = 10000;
-      FD_ZERO(&readfds);
-      FD_SET(fd, &readfds);
-      FD_ZERO(&exceptfds);
-      FD_SET(fd, &exceptfds);
-      select(1, &readfds, 0, &exceptfds, &tv);
-#else
-    #ifdef __APPLE__
-      struct pollfd pfd = { fd, POLLIN, 0 };
-    #else
-      struct pollfd pfd = { fd, POLLIN|POLLRDHUP, 0 };
-    #endif
-      poll(&pfd, 1, 10);
-#endif
-      rc = _amps_SSL_shutdown(ssl);
-    }
-  }
-  if (fd != AMPS_INVALID_SOCKET)
-  {
-    shutdown(fd, SHUT_RDWR);
-    AMPS_CLOSESOCKET(fd);
-  }
-  if (ssl)
-  {
-    _amps_SSL_free(ssl);
-  }
-  return rc;
-}
-
-// Pointer to function for closing SSL/fd that Python client redirects
-AMPSDLL int (*amps_tcps_close_all)(AMPS_SOCKET, _amps_SSL*) = _amps_tcps_close_all;
-
-const amps_char* amps_tcps_get_last_error(
-  amps_handle transport)
-{
-  return ((amps_tcps_t*)transport)->lastErrorBuf;
-}
-
-void amps_tcps_clear_error(
-  amps_handle transport)
-{
-  ((amps_tcps_t*)transport)->lastErrorBuf[0] = '\0';
+  _AMPS_SNPRINTF(transport->lastErrorBuf, sizeof(transport->lastErrorBuf),
+                 "zlib error code %d: %s", rc, transport->pDeflater->msg);
 }
 
 void amps_tcps_set_error(
@@ -374,7 +343,6 @@ void amps_tcps_set_error(
   const amps_char* buffer)
 {
   _AMPS_SNPRINTF(transport->lastErrorBuf, sizeof(transport->lastErrorBuf), "%s", buffer);
-  transport->lastErrorBuf[sizeof(transport->lastErrorBuf) - 1] = '\0';
 }
 
 // Allow python client to set transport errors during connect/close
@@ -458,6 +426,216 @@ void amps_tcps_set_ssl_error(amps_tcps_t* transport,
   }
 }
 
+int amps_tcps_send_compressed(amps_tcps_t* me, amps_char* data, size_t length)
+{
+  amps_zstream* pDeflater = me->pDeflater;
+
+  pDeflater->next_in = data;
+  pDeflater->avail_in = (unsigned int)length;
+  pDeflater->avail_out = (unsigned int)me->deflateBufferLength;
+
+  // The ... || !pDeflater->avail_out is so we can send the flushed bytes
+  // even if all the input data is consumed, when the output buffer fills up.
+  while (pDeflater->avail_in || !pDeflater->avail_out)
+  {
+    // We send the whole deflate buffer after every deflate().
+    pDeflater->avail_out = (unsigned int)me->deflateBufferLength;
+    pDeflater->next_out = me->pDeflateBuffer;
+
+    int rc = amps_deflate(pDeflater, 2);
+    if (rc != 0)
+    {
+      amps_tcps_set_zlib_error(me, rc);
+      return -1;
+    }
+
+    // Send the whole deflate buffer here.
+    const char* p = me->pDeflateBuffer, *pe = pDeflater->next_out;
+    while (p != pe)
+    {
+      int sentBytes = _amps_SSL_write(me->ssl, p, (int)(pe - p));
+      if (sentBytes <= 0)
+      {
+        return -1;
+      }
+      p += sentBytes;
+    }
+  }
+  return (int)length;
+}
+
+int amps_tcps_recv_compressed(amps_tcps_t* me, size_t capacity)
+{
+  amps_zstream* pInflater = me->pInflater;
+  int bytes = 0;
+  if (!pInflater->avail_in && !me->inflatePending)
+  {
+    // Retrieve data from the socket only if there's nothing to do.
+    errno = 0;
+    bytes = _amps_SSL_read(me->ssl, me->pInflateBuffer, (int)me->inflateBufferLength);
+    if (bytes <= 0)
+    {
+      return bytes;
+    }
+    pInflater->avail_in = (unsigned int)bytes;
+    pInflater->next_in = me->pInflateBuffer;
+  }
+  pInflater->avail_out = (unsigned int)capacity;
+
+  if (pInflater->avail_out)
+  {
+    me->inflatePending = 0;
+    int rc = amps_inflate(pInflater, 2);
+    if (rc != 0)
+    {
+      amps_tcps_set_zlib_error(me, rc);
+      if (rc > 0)
+      {
+        return -1;
+      }
+      return rc;
+    }
+    if (pInflater->avail_out == 0)
+    {
+      // Special case where, even though we consumed the input, we have more output to write next time someone calls here.
+      me->inflatePending = 1;
+      // Are we out of space to inflate?
+      if (pInflater->avail_in
+          && pInflater->next_in < (me->pInflateBuffer + bytes))
+      {
+        return AMPS_ZLIB_NEED_SPACE;
+      }
+    }
+    // If we inflated nothing, read more input
+    if (pInflater->avail_out == capacity)
+    {
+      return AMPS_ZLIB_WANT_READ;
+    }
+    return (int)(capacity - pInflater->avail_out);
+  }
+
+  return 0;
+}
+
+amps_result
+amps_tcps_init_zlib(amps_tcps_t* me)
+{
+  const int windowBits = -15; /* no headers */
+  if (me->pDeflater)
+  {
+    amps_deflateEnd(me->pDeflater);
+    free(me->pDeflater);
+    me->pDeflater = 0;
+
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
+  }
+  me->pDeflater = (amps_zstream*)malloc(sizeof(amps_zstream));
+  memset(me->pDeflater, 0, sizeof(amps_zstream));
+  int rc = amps_deflateInit2(me->pDeflater,
+                  6,   /* level */
+                  8,   /* Z_DEFLATED */
+                  windowBits,
+                  8,   /* default memory level */
+                  0);  /* default strategy */
+  if (rc != 0)
+  {
+    free(me->pDeflater);
+    me->pDeflater = 0;
+    amps_tcps_set_error(me, "Unable to initialize zlib.");
+    return AMPS_E_URI;
+  }
+  me->pInflater = (amps_zstream*)malloc(sizeof(amps_zstream));
+  memset(me->pInflater, 0, sizeof(amps_zstream));
+  rc = amps_inflateInit2(me->pInflater, windowBits)
+  if (rc != 0)
+  {
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
+    amps_tcps_set_error(me, "Unable to initialize zlib.");
+    return AMPS_E_URI;
+  }
+
+  //  Initialize the work buffers for inflation and deflation;
+  if (!me->pDeflateBuffer)
+  {
+    me->pDeflateBuffer = (char*)malloc(16384);
+    me->deflateBufferLength = 16384;
+  }
+
+  if (!me->pInflateBuffer)
+  {
+    me->pInflateBuffer = (char*)malloc(16384);
+    me->inflateBufferLength = 16384;
+  }
+  me->inflatePending = 0;
+#if __STDC_VERSION__ >= 201100
+  me->useZlib = true;
+#else
+  AMPS_IEX_LONG(&me->useZlib, 1);
+#endif
+  return AMPS_E_OK;
+}
+
+// Function to close SSL* and fd that can be replaced by Python client
+int _amps_tcps_close_all(AMPS_SOCKET fd, _amps_SSL* ssl)
+{
+  int rc = 0;
+  if (ssl)
+  {
+    rc = _amps_SSL_shutdown(ssl);
+    if (rc == 0 && fd != AMPS_INVALID_SOCKET)
+    {
+      // attempt to wait for the server to send back the matching close notify.
+#ifdef _WIN32
+      struct timeval tv;
+      fd_set readfds, exceptfds;
+      tv.tv_sec = 0;
+      tv.tv_usec = 10000;
+      FD_ZERO(&readfds);
+      FD_SET(fd, &readfds);
+      FD_ZERO(&exceptfds);
+      FD_SET(fd, &exceptfds);
+      select(1, &readfds, 0, &exceptfds, &tv);
+#else
+    #ifdef __APPLE__
+      struct pollfd pfd = { fd, POLLIN, 0 };
+    #else
+      struct pollfd pfd = { fd, POLLIN|POLLRDHUP, 0 };
+    #endif
+      poll(&pfd, 1, 10);
+#endif
+    }
+  }
+  if (fd != AMPS_INVALID_SOCKET)
+  {
+    shutdown(fd, SHUT_RDWR);
+    AMPS_CLOSESOCKET(fd);
+  }
+  if (ssl)
+  {
+    _amps_SSL_free(ssl);
+  }
+  return rc;
+}
+
+// Pointer to function for closing SSL/fd that Python client redirects
+AMPSDLL int (*amps_tcps_close_all)(AMPS_SOCKET, _amps_SSL*) = _amps_tcps_close_all;
+
+const amps_char* amps_tcps_get_last_error(
+  amps_handle transport)
+{
+  return ((amps_tcps_t*)transport)->lastErrorBuf;
+}
+
+void amps_tcps_clear_error(
+  amps_handle transport)
+{
+  ((amps_tcps_t*)transport)->lastErrorBuf[0] = '\0';
+}
+
 void
 amps_tcps_noop_filter_function(const unsigned char* data, size_t length, short direction, void* userdata)
 {;}
@@ -491,6 +669,9 @@ amps_handle amps_tcps_create(void)
 #if __STDC_VERSION__ >= 201100
     transport->disconnecting = false;
     transport->destroying = false;
+#else
+    AMPS_IEX_LONG(&transport->disconnecting, 0);
+    AMPS_IEX_LONG(&transport->destroying, 0);
 #endif
     transport->filterFunction = &amps_tcps_noop_filter_function;
     amps_atfork_add(transport, (_amps_atfork_callback_function)amps_tcps_atfork_handler);
@@ -539,9 +720,10 @@ amps_tcps_opt_parse(const char* val, size_t valLen, int* parsed)
 
 typedef enum
 {
-  AMPS_PROPERTY_OK,
-  AMPS_PROPERTY_PREFLIGHT_TRUE,
-  AMPS_PROPERTY_PARSE_ERROR
+  AMPS_PROPERTY_PARSE_ERROR = 0,
+  AMPS_PROPERTY_OK = 1,
+  AMPS_PROPERTY_PREFLIGHT_TRUE = 2,
+  AMPS_PROPERTY_COMPRESSION_ZLIB = 4,
 } amps_property_result;
 
 amps_property_result
@@ -659,6 +841,12 @@ amps_tcps_apply_socket_property(AMPS_SOCKET fd, const char* key, size_t keyLen,
     // no-op; handled in amps_tcps_apply_ssl_properties
     return AMPS_PROPERTY_OK;
   }
+  else if (keyLen == 11 && memcmp(key, "compression", keyLen) == 0)
+  {
+    return (valLen == 4 && memcmp(val, "zlib", 4) == 0)
+            ? AMPS_PROPERTY_COMPRESSION_ZLIB
+            : AMPS_PROPERTY_PARSE_ERROR;
+  }
   // properties after this point must only be boolean or numeric
   if (!amps_tcps_opt_parse(val, valLen, &value))
   {
@@ -715,7 +903,7 @@ amps_property_result
 amps_tcps_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
                                   size_t uriLength, amps_uri_state* uriState)
 {
-  amps_property_result retVal = AMPS_PROPERTY_OK;
+  int retVal = AMPS_PROPERTY_OK;
   amps_property_result optRet = AMPS_PROPERTY_OK;
   const char* key = NULL;
   size_t keyLength = 0;
@@ -746,7 +934,7 @@ amps_tcps_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
         {
           return optRet;
         }
-        retVal = optRet;
+        retVal |= (int)optRet;
       }
     }
   }
@@ -754,7 +942,7 @@ amps_tcps_apply_socket_properties(AMPS_SOCKET fd, const char* uri,
   {
     return AMPS_PROPERTY_PARSE_ERROR;
   }
-  return retVal;
+  return (amps_property_result)retVal;
 }
 
 // Retrieve server name indicator for use by SSL when connecting
@@ -797,6 +985,8 @@ _amps_SSL* _amps_tcps_get_socket_and_ssl(AMPS_SOCKET fd_,
   {
     if (amps_ssl_init(NULL))
     {
+      amps_tcps_set_error(transport, amps_ssl_get_error());
+      AMPS_CLOSESOCKET(fd_);
       return NULL;
     }
   }
@@ -813,6 +1003,7 @@ _amps_SSL* _amps_tcps_get_socket_and_ssl(AMPS_SOCKET fd_,
     _amps_SSL_free(ssl);
     ssl = (_amps_SSL*)0;
     transport->ssl = (_amps_SSL*)0;
+    AMPS_CLOSESOCKET(fd_);
     return NULL;
   }
 
@@ -824,6 +1015,7 @@ _amps_SSL* _amps_tcps_get_socket_and_ssl(AMPS_SOCKET fd_,
   {
     amps_tcps_set_socket_error(transport, sysError);
     _amps_SSL_free(ssl);
+    AMPS_CLOSESOCKET(fd_);
     return NULL;
   }
 
@@ -837,6 +1029,7 @@ _amps_SSL* _amps_tcps_get_socket_and_ssl(AMPS_SOCKET fd_,
     {
       amps_tcps_set_ssl_error(transport, rc, sysError);
       _amps_SSL_free(ssl);
+      AMPS_CLOSESOCKET(fd_);
       return NULL;
     }
   }
@@ -850,6 +1043,7 @@ _amps_SSL* _amps_tcps_get_socket_and_ssl(AMPS_SOCKET fd_,
   {
     amps_tcps_set_ssl_error(transport, rc, sysError);
     _amps_SSL_free(ssl);
+    AMPS_CLOSESOCKET(fd_);
     return NULL;
   }
   return ssl;
@@ -868,18 +1062,21 @@ int amps_tcps_locked_read(amps_tcps_t*, char*, int, int*);
 amps_result
 amps_tcps_update_read_timeout(amps_tcps_t* me)
 {
-  int timeout = (int)((me->readTimeoutMillis && me->idleTimeMillis) ?
-                      AMPS_MIN(me->readTimeoutMillis, me->idleTimeMillis) :
-                      AMPS_MAX(me->readTimeoutMillis, me->idleTimeMillis));
+  amps_uint64_t timeout = (me->readTimeoutMillis && me->idleTimeMillis)
+                          ? AMPS_MIN(me->readTimeoutMillis, me->idleTimeMillis)
+                          : AMPS_MAX(me->readTimeoutMillis, me->idleTimeMillis);
+  timeout = (timeout && me->sendBatchTimeoutMillis)
+            ? AMPS_MIN(timeout, me->sendBatchTimeoutMillis)
+            : AMPS_MAX(timeout, me->sendBatchTimeoutMillis);
   int rc = 0;
 #ifdef _WIN32
-  DWORD timeoutMillis = timeout;
+  DWORD timeoutMillis = (DWORD)timeout;
   rc = setsockopt(me->fd, SOL_SOCKET, SO_RCVTIMEO,
                   (const char*)&timeoutMillis, sizeof(DWORD));
 #else
   struct timeval tv;
-  tv.tv_sec = timeout / 1000;
-  tv.tv_usec = (timeout % 1000) * 1000;
+  tv.tv_sec = (int)timeout / 1000;
+  tv.tv_usec = ((int)timeout % 1000) * 1000;
   rc = setsockopt(me->fd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(struct timeval));
 #endif
   int sysError = GET_ERRNO;
@@ -911,6 +1108,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
   /*now, send */
   while (socketBytes < bytesWritten)
   {
+    errno = 0;
     int bytesWrittenThisTime;
     bytesWrittenThisTime = _amps_SSL_write(me->ssl, httpPreflight + socketBytes, bytesWritten - socketBytes);
     if (bytesWrittenThisTime <= 0)
@@ -934,6 +1132,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
     // Use select to wait for data if none already exists on the ssl.
     if (!_amps_SSL_pending(me->ssl))
     {
+      errno = 0;
   #ifdef _WIN32
       struct timeval tv;
       fd_set readfds, exceptfds;
@@ -950,7 +1149,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
     #else
       struct pollfd pfd = { me->fd, POLLIN|POLLRDHUP, 0 };
     #endif
-      socketBytes = poll(&pfd, 1, 1);
+      socketBytes = poll(&pfd, 1, 100);
       if (socketBytes > 0)
       {
         // If there are no bytes ready to read, we lost connection
@@ -969,6 +1168,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
   #endif
       if (socketBytes > 0)
       {
+          errno = 0;
           socketBytes = _amps_SSL_read(me->ssl, readPoint, (int)(end - readPoint));
           errorcode = socketBytes <= 0 ? _amps_SSL_get_error(me->ssl, socketBytes) : 0;
       }
@@ -1019,6 +1219,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
     // Use select to wait for data if none already exists on the ssl.
     if (!_amps_SSL_pending(me->ssl))
     {
+      errno = 0;
   #ifdef _WIN32
       struct timeval tv;
       fd_set readfds, exceptfds;
@@ -1035,7 +1236,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
     #else
       struct pollfd pfd = { me->fd, POLLIN|POLLRDHUP, 0 };
     #endif
-      socketBytes = poll(&pfd, 1, 1);
+      socketBytes = poll(&pfd, 1, 100);
       if (socketBytes > 0)
       {
         // If there are no bytes ready to read, we lost connection
@@ -1054,6 +1255,7 @@ amps_result _amps_tcps_handle_http_preflight(amps_tcps_t* me)
   #endif
       if (socketBytes > 0)
       {
+          errno = 0;
           socketBytes = _amps_SSL_read(me->ssl, readPoint, (int)(end - readPoint));
           errorcode = socketBytes <= 0 ? _amps_SSL_get_error(me->ssl, socketBytes) : 0;
       }
@@ -1124,8 +1326,13 @@ amps_result amps_tcps_connect(amps_handle transport, const amps_char* address)
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me->lock);
 #endif
+  me->lastErrorBuf[0] = '\0';
   /* if we were previously marked as disconnecting, forget that. */
+#if __STDC_VERSION__ >= 201100
+    me->disconnecting = false;
+#else
   AMPS_IEX_LONG(&me->disconnecting, 0);
+#endif
   AMPS_LOCK(&me->sendLock);
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me->sendLock);
@@ -1257,32 +1464,77 @@ amps_result amps_tcps_connect(amps_handle transport, const amps_char* address)
     result = AMPS_E_CONNECTION_REFUSED;
     goto error;
   }
-  AMPS_IEX(&me->fd, fd);
 
   /* apply socket properties */
-  propResult = amps_tcps_apply_socket_properties(me->fd, address, address_length, &uri_state_socket_props);
+  propResult = amps_tcps_apply_socket_properties(fd, address, address_length, &uri_state_socket_props);
   if (me->destroying || propResult == AMPS_PROPERTY_PARSE_ERROR)
   {
     freeaddrinfo(pAddrInfo);
-    if (!me->destroying)
+    if (me->lastErrorBuf[0] == '\0' && propResult == AMPS_PROPERTY_PARSE_ERROR)
     {
-      fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
+      char errBuf[128];
+      _AMPS_SNPRINTF(errBuf, 128, "The URI specified invalid properties: %.*s.", (int)uri_state_socket_props.part_length, uri_state_socket_props.part);
+      amps_tcps_set_error(me, errBuf);
     }
-    amps_tcps_set_error(me, "The URI specified invalid properties.");
+    AMPS_CLOSESOCKET(fd);
+    fd = AMPS_INVALID_SOCKET;
     result = AMPS_E_URI;
     goto error;
   }
+  if (propResult & AMPS_PROPERTY_COMPRESSION_ZLIB)
+  {
+    if (amps_zlib_init(NULL) == -1)
+    {
+      amps_tcps_set_error(me, amps_zlib_last_error);
+      freeaddrinfo(pAddrInfo);
+      result = AMPS_E_URI;
+#if __STDC_VERSION__ >= 201100
+      me->useZlib = false;
+#else
+      AMPS_IEX_LONG(&me->useZlib, 0);
+#endif
+      AMPS_CLOSESOCKET(fd);
+      fd = AMPS_INVALID_SOCKET;
+      result = AMPS_E_URI;
+      goto error;
+    }
+    if (amps_tcps_init_zlib(me) != AMPS_E_OK)
+    {
+      freeaddrinfo(pAddrInfo);
+      result = AMPS_E_URI;
+#if __STDC_VERSION__ >= 201100
+      me->useZlib = false;
+#else
+      AMPS_IEX_LONG(&me->useZlib, 0);
+#endif
+      AMPS_CLOSESOCKET(fd);
+      fd = AMPS_INVALID_SOCKET;
+      result = AMPS_E_URI;
+      goto error;
+    }
+#if __STDC_VERSION__ >= 201100
+    me->useZlib = true;
+#else
+    AMPS_IEX_LONG(&me->useZlib, 1);
+#endif
+  }
+  else
+  {
+#if __STDC_VERSION__ >= 201100
+    me->useZlib = false;
+#else
+    AMPS_IEX_LONG(&me->useZlib, 0);
+#endif
+  }
 #ifdef __APPLE__
   rc = 1; /* Set NOSIGPIPE to ON */
   if (me->destroying
-      || setsockopt(me->fd, SOL_SOCKET, SO_NOSIGPIPE, &rc, sizeof(rc)) < 0)
+      || setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &rc, sizeof(rc)) < 0)
   {
     freeaddrinfo(pAddrInfo);
-    if (!me->destroying)
-    {
-      fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
-    }
     amps_tcps_set_error(me, "Failed to set no SIGPIPE.");
+    AMPS_CLOSESOCKET(fd);
+    fd = AMPS_INVALID_SOCKET;
     result = AMPS_E_CONNECTION;
     goto error;
   }
@@ -1292,14 +1544,10 @@ amps_result amps_tcps_connect(amps_handle transport, const amps_char* address)
   char* hostname = amps_tcps_get_sni_hostname(address, address_length, &uri_state_ssl_props);
   int portNum = atoi(port);
   amps_tcps_clear_error(me);
-  me->ssl = amps_tcps_get_socket_and_ssl((AMPS_SOCKET)(me->fd), host, portNum, pAddrInfo, hostname, (void*)me);
+  me->ssl = amps_tcps_get_socket_and_ssl(fd, host, portNum, pAddrInfo, hostname, (void*)me);
   freeaddrinfo(pAddrInfo);
   if (me->destroying || me->ssl == 0)
   {
-    if (!me->destroying)
-    {
-      fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
-    }
     if (!amps_tcps_get_last_error(me))
     {
       amps_tcps_set_error_stack_error(me);
@@ -1309,9 +1557,12 @@ amps_result amps_tcps_connect(amps_handle transport, const amps_char* address)
         amps_tcps_set_error(me, "Could not initialize an SSL context or SSL.");
       }
     }
+    fd = AMPS_INVALID_SOCKET;
     result = AMPS_E_SSL;
     goto error;
   }
+  AMPS_IEX(&me->fd, fd);
+  fd = AMPS_INVALID_SOCKET;
 
   // tcps transport must use read with a timeout.
   amps_tcps_set_idle_time(transport, 100);
@@ -1322,7 +1573,7 @@ amps_result amps_tcps_connect(amps_handle transport, const amps_char* address)
   result = AMPS_E_OK;
 
   // Handle http_preflight upgrade from HTTP
-  if (propResult == AMPS_PROPERTY_PREFLIGHT_TRUE && me->httpPreflightCallback)
+  if (propResult & AMPS_PROPERTY_PREFLIGHT_TRUE && me->httpPreflightCallback)
   {
     result = _amps_tcps_handle_http_preflight(me);
     if (result != AMPS_E_OK)
@@ -1396,11 +1647,8 @@ error:
   if (result != AMPS_E_OK)
   {
     // Use our close function to shutdown socket and SSL
-    int rc = amps_tcps_close_all(fd, ssl);
-    if (rc < 0)
-    {
-      amps_tcps_set_ssl_error(me, rc, sysError);
-    }
+    // But don't worry about return value as error state is already set
+    amps_tcps_close_all(fd, ssl);
 #ifdef _WIN32
     me->thread = 0;
 #else
@@ -1452,7 +1700,11 @@ void amps_tcps_close(amps_handle transport)
   amps_tcps_t* me = (amps_tcps_t*)transport;
   AMPS_SOCKET fd = AMPS_INVALID_SOCKET;
   _amps_SSL* ssl = 0;
+#if __STDC_VERSION__ >= 201100
+  me->disconnecting = true;
+#else
   AMPS_IEX_LONG(&me->disconnecting, 1);
+#endif
   AMPS_LOCK(&me->sendLock);
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me->sendLock);
@@ -1524,8 +1776,25 @@ void amps_tcps_destroy(amps_handle transport)
   AMPS_JOIN(me->thread);
   AMPS_SLEEP(1);
   free(me->buf);
-  /* Hopefully, nobody else is using me right now. */
+  if(me->pDeflater)
+  {
+    amps_deflateEnd(me->pDeflater);
+    free(me->pDeflater);
+    me->pDeflater = 0;
+    free(me->pDeflateBuffer);
+    me->pDeflateBuffer = 0;
+  }
+  if(me->pInflater)
+  {
+    amps_inflateEnd(me->pInflater);
+    free(me->pInflater);
+    me->pInflater = 0;
 
+    free(me->pInflateBuffer);
+    me->pInflateBuffer = 0;
+  }
+  free(me->sendCache);
+  /* Hopefully, nobody else is using me right now. */
   AMPS_KILLLOCK(&me->lock);
   AMPS_KILLLOCK(&me->sendLock);
 
@@ -1547,6 +1816,7 @@ amps_result amps_tcps_handle_disconnect(
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me->sendLock);
 #endif
+  me->inflatePending = 0;
   fd = AMPS_IEX_GET(&me->fd, AMPS_INVALID_SOCKET);
   ssl = (_amps_SSL*)AMPS_IEX_GET(&me->ssl, 0);
   AMPS_UNLOCK(&me->sendLock);
@@ -1633,13 +1903,74 @@ void amps_tcps_handle_stream_corruption(
   amps_tcps_handle_disconnect(me, failedConnectionVersion);
 }
 
+int
+amps_tcps_send_bytes(amps_tcps_t* me, char* buf, int bytesWritten)
+{
+  int bytesSent = 0;
+  while (bytesSent < bytesWritten)
+  {
+    int bytesWrittenThisTime;
+    if (me->useZlib)
+    {
+      bytesWrittenThisTime = amps_tcps_send_compressed(me, buf + bytesSent, (size_t)(bytesWritten - bytesSent));
+    }
+    else
+    {
+      bytesWrittenThisTime = _amps_SSL_write(me->ssl, buf + bytesSent, bytesWritten - bytesSent);
+    }
+    if (bytesWrittenThisTime <= 0)
+    {
+      // If error code is want read or want write, this is retryable
+      int errorcode_ = _amps_SSL_get_error(me->ssl, bytesWrittenThisTime);
+      if (errorcode_ == AMPS_SSL_ERROR_WANT_READ
+          || errorcode_ == AMPS_SSL_ERROR_WANT_WRITE)
+      {
+        AMPS_UNLOCK(&me->sendLock);
+        if (me->disconnecting)
+        {
+          amps_tcps_set_error(me, "Disconnecting.");
+          return bytesWrittenThisTime;
+        }
+        if (me->fd == AMPS_INVALID_SOCKET)
+        {
+          amps_tcps_set_error(me, "Not connected.");
+          return bytesWrittenThisTime;
+        }
+        AMPS_LOCK(&me->sendLock);
+        if (me->ssl == 0)
+        {
+          AMPS_UNLOCK(&me->sendLock);
+          amps_tcps_set_error(me, "Not connected.");
+          return bytesWrittenThisTime;
+        }
+        continue;
+      }
+      return bytesWrittenThisTime;
+    }
+    bytesSent += bytesWrittenThisTime;
+  }
+  return bytesSent;
+}
+
+int
+amps_tcps_send_cache(amps_tcps_t* me)
+{
+  me->lastSendTime = amps_now();
+  int bytesSent = amps_tcps_send_bytes(me, me->sendCache, (int)(me->sendCacheOffset));
+  if (bytesSent >= 0)
+  {
+      me->sendCacheOffset = 0;
+  }
+  return bytesSent;
+}
 
 amps_result
-amps_tcps_send_with_version(amps_handle transport,
-                            amps_handle message,
-                            unsigned* version_out)
+amps_tcps_send_batch(amps_handle transport,
+                     amps_handle message,
+                     unsigned* version_out,
+                     int addToBatch)
 {
-  amps_result result = AMPS_E_OK;
+  amps_result result = AMPS_E_DISCONNECTED;
   amps_tcps_t* me = (amps_tcps_t*)transport;
   size_t len = AMPS_TCPS_DEFAULT_BUF_LEN;
   int bytesWritten = -1;
@@ -1668,40 +1999,122 @@ amps_tcps_send_with_version(amps_handle transport,
   if (me->ssl == 0)
   {
     amps_tcps_set_error(me, "Not connected.");
-    result = AMPS_E_CONNECTION;
+    result = AMPS_E_DISCONNECTED;
     goto error;
   }
   while (bytesWritten < 0)
   {
-    if (me->buf == NULL)
+    if (me->sendCache && addToBatch == 0)
     {
-      me->buf = (char*)malloc(len);
-      if (me->buf == NULL)
+      // Send the cache
+      bytesSent = amps_tcps_send_cache(me);
+      if (bytesSent < 0)
       {
-        amps_tcps_set_error(me, "Unable to allocate memory to send message.");
-        result = AMPS_E_MEMORY;
+          AMPS_UNLOCK(&me->sendLock);
+          result = AMPS_E_DISCONNECTED;
+          goto error;
+      }
+    }
+    if (addToBatch && me->sendBatchSizeBytes)
+    {
+      // Serialize into cache
+      if (me->sendCache == NULL)
+      {
+        me->sendCache = (char*)malloc(me->sendBatchSizeBytes);
+        if (me->sendCache == NULL)
+        {
+          amps_tcps_set_error(me, "Unable to allocate memory to cache message.");
+          AMPS_UNLOCK(&me->sendLock);
+          result = AMPS_E_MEMORY;
+          goto error;
+        }
+      }
+      bytesWritten = amps_message_serialize(message, me->serializer,
+                                            (me->sendCache) + me->sendCacheOffset + 4,
+                                            me->sendBatchSizeBytes - me->sendCacheOffset - 4);
+      /* amps_message_serialize could have written 0 bytes.
+       * who are we to judge?  If it returns less than 0,
+       * there wasn't enough room. */
+      if (bytesWritten >= 0)
+      {
+        // record the message length all up in the first 4 bytes.
+        bytesWrittenN = htonl((unsigned int)bytesWritten);
+        me->filterFunction((const unsigned char*)(me->sendCache + me->sendCacheOffset + 4),
+                           (size_t)bytesWritten, 0, me->filterUserData);
+        *((unsigned int*)(me->sendCache + me->sendCacheOffset)) = bytesWrittenN;
+        bytesWritten += 4;
+        if (!me->sendCacheOffset)
+        {
+          // Start the send clock based on first cachec message
+          me->lastSendTime = amps_now();
+        }
+        me->sendCacheOffset += (size_t)bytesWritten;
+        if (me->sendCacheOffset < (me->sendBatchSizeBytes - 32)
+            && me->lastSendTime > (amps_now() - me->sendBatchTimeoutMillis))
+        {
+          result = AMPS_E_OK;
+          // Not an error, but that's where cleanup is
+          goto error;
+        }
+        bytesSent = amps_tcps_send_cache(me);
+        if (bytesSent < 0)
+        {
+          result = AMPS_E_DISCONNECTED;
+          goto error;
+        }
+        result = AMPS_E_OK;
+        // Not an error, but that's where cleanup is
         goto error;
       }
-      me->capacity = len;
+      // Need more space, bytesWritten < 0
+      if (me->sendCacheOffset)
+      {
+        // Send cache and try again
+        bytesSent = amps_tcps_send_cache(me);
+        if (bytesSent < 0)
+        {
+          result = AMPS_E_DISCONNECTED;
+          goto error;
+        }
+      }
+      else
+      {
+        // Message is larger than batch max bytes, so just send it
+        addToBatch = 0;
+      }
     }
     else
     {
-      len = me->capacity;
+      if (me->buf == NULL)
+      {
+        me->buf = (char*)malloc(len);
+        if (me->buf == NULL)
+        {
+          amps_tcps_set_error(me, "Unable to allocate memory to send message.");
+          result = AMPS_E_MEMORY;
+          goto error;
+        }
+        me->capacity = len;
+      }
+      else
+      {
+        len = me->capacity;
+      }
+      /* reserve 4 bytes for the length of the message */
+      bytesWritten = amps_message_serialize(message, me->serializer, (me->buf) + 4, len - 4);
+      /* amps_message_serialize could have written 0 bytes.
+       * who are we to judge?  If it returns less than 0,
+       * there wasn't enough room. */
+      if (bytesWritten >= 0)
+      {
+        break;
+      }
+      /* free this buffer, allocate a larger buffer next time */
+      free(me->buf);
+      me->capacity = 0;
+      me->buf = NULL;
+      len = (size_t)((double)len * 1.5);
     }
-    /* reserve 4 bytes for the length of the message */
-    bytesWritten = amps_message_serialize(message, me->serializer, (me->buf) + 4, len - 4);
-    /* amps_message_serialize could have written 0 bytes.
-     * who are we to judge?  If it returns less than 0,
-     * there wasn't enough room. */
-    if (bytesWritten >= 0)
-    {
-      break;
-    }
-    /* free this buffer, allocate a larger buffer next time */
-    free(me->buf);
-    me->capacity = 0;
-    me->buf = NULL;
-    len = (size_t)((double)len * 1.5);
   }
   /* once we're done, we don't free buf -- it hangs around until we need it next time.
    * record the message length all up in the first 4 bytes. */
@@ -1710,48 +2123,18 @@ amps_tcps_send_with_version(amps_handle transport,
   *((unsigned int*)(me->buf)) = bytesWrittenN;
   bytesWritten += 4;
   /*now, send */
-  while (bytesSent < bytesWritten)
+  bytesSent = amps_tcps_send_bytes(me, me->buf, bytesWritten);
+  if (bytesSent <= 0)
   {
-    int bytesWrittenThisTime;
-    bytesWrittenThisTime = _amps_SSL_write(me->ssl, (me->buf) + bytesSent, bytesWritten - bytesSent);
-    if (bytesWrittenThisTime <= 0)
+    /* record the error */
+    if (me->lastErrorBuf[0] == '\0')
     {
-      // If error code is want read or want write, this is retryable
-      int errorcode_ = _amps_SSL_get_error(me->ssl, bytesWrittenThisTime);
-      if (errorcode_ == AMPS_SSL_ERROR_WANT_READ
-          || errorcode_ == AMPS_SSL_ERROR_WANT_WRITE)
-      {
-        AMPS_UNLOCK(&me->sendLock);
-        if (me->disconnecting)
-        {
-          amps_tcps_set_error(me, "Disconnecting.");
-          result = AMPS_E_DISCONNECTED;
-          AMPS_LOCK(&me->sendLock);
-          goto error;
-        }
-        if (me->fd == AMPS_INVALID_SOCKET)
-        {
-          amps_tcps_set_error(me, "Not connected.");
-          result = AMPS_E_DISCONNECTED;
-          AMPS_LOCK(&me->sendLock);
-          goto error;
-        }
-        AMPS_LOCK(&me->sendLock);
-        if (me->ssl == 0)
-        {
-          amps_tcps_set_error(me, "Not connected.");
-          result = AMPS_E_CONNECTION;
-          goto error;
-        }
-        continue;
-      }
-      /* record the error */
       amps_tcps_set_error(me, "The connection is closed.");
-      result = AMPS_E_DISCONNECTED;
-      goto error;
     }
-    bytesSent += bytesWrittenThisTime;
+    result = AMPS_E_DISCONNECTED;
+    goto error;
   }
+  result = AMPS_E_OK;
 error:
   AMPS_UNLOCK(&me->sendLock);
 #ifndef _WIN32
@@ -1761,12 +2144,20 @@ error:
   return result;
 }
 
+amps_result
+amps_tcps_send_with_version(amps_handle transport,
+                            amps_handle message,
+                            unsigned* version_out)
+{
+  return amps_tcps_send_batch(transport, message, version_out, 0);
+}
+
 amps_result
 amps_tcps_send(amps_handle transport,
                amps_handle message)
 {
   unsigned version_out;
-  return amps_tcps_send_with_version(transport, message, &version_out);
+  return amps_tcps_send_batch(transport, message, &version_out, 0);
 }
 
 /* Returns # bytes read, 0 if a timeout occurred without a read,
@@ -1811,6 +2202,7 @@ int amps_tcps_locked_read(amps_tcps_t* me_,
     FD_ZERO(&exceptfds);
     FD_SET(fd, &exceptfds);
     AMPS_UNLOCK(&me_->sendLock);
+    errno = 0;
     rc  = select(1, &readfds, 0, &exceptfds, &tv);
 #else
   #ifdef __APPLE__
@@ -1819,7 +2211,8 @@ int amps_tcps_locked_read(amps_tcps_t* me_,
     struct pollfd pfd = { fd, POLLIN|POLLRDHUP, 0 };
   #endif
     AMPS_UNLOCK(&me_->sendLock);
-    rc = poll(&pfd, 1, 1);
+    errno = 0;
+    rc = poll(&pfd, 1, 100);
     if (rc > 0)
     {
       // If there are no bytes ready to read, we lost connection
@@ -1829,7 +2222,7 @@ int amps_tcps_locked_read(amps_tcps_t* me_,
       }
     }
 #endif
-    if (rc <= 0)
+    if (rc < 0)
     {
       return rc;
     }
@@ -1846,8 +2239,25 @@ int amps_tcps_locked_read(amps_tcps_t* me_,
 #ifndef _WIN32
   pthread_cleanup_push(amps_cleanup_unlock_mutex, (void*)&me_->sendLock);
 #endif
-  rc = _amps_SSL_read(me_->ssl, readPoint_, bytes_);
-  *pErrorcode_ = rc <= 0 ? _amps_SSL_get_error(me_->ssl, rc) : 0;
+  if (me_->useZlib
+      && (rc > 0
+          || (rc == 0
+              && me_->inflatePending)))
+  {
+    // Make sure next_out is set
+    if (!me_->pInflater->next_out)
+    {
+      me_->pInflater->next_out = (const char*)readPoint_;
+    }
+    rc = amps_tcps_recv_compressed(me_, (size_t)((const char*)(readPoint_ + bytes_) - me_->pInflater->next_out));
+    *pErrorcode_ = rc <= 0 ? _amps_SSL_get_error(me_->ssl, rc) : 0;
+  }
+  else if (rc > 0)
+  {
+    errno = 0;
+    rc = _amps_SSL_read(me_->ssl, readPoint_, bytes_);
+    *pErrorcode_ = rc <= 0 ? _amps_SSL_get_error(me_->ssl, rc) : 0;
+  }
 
   AMPS_UNLOCK(&me_->sendLock);
 #ifndef _WIN32
@@ -1861,6 +2271,11 @@ int amps_tcps_locked_read(amps_tcps_t* me_,
       *pErrorcode_ = 0;
       return 0;
     }
+    if (rc == AMPS_ZLIB_WANT_READ
+        || rc == AMPS_ZLIB_NEED_SPACE)
+    {
+      return rc;
+    }
     // End of stream
     return -1;
   }
@@ -1924,6 +2339,7 @@ void* amps_tcps_threaded_reader(void* userData)
   int cancelState = 0;
   const size_t BUFFER_SIZE = AMPS_TCPS_DEFAULT_BUF_LEN;
   const unsigned MAX_MESSAGE_SIZE = 1024 * 1024 * 1024;
+  amps_uint64_t now = 0;
   amps_uint64_t lastReadTime = 0;
   amps_uint64_t lastIdleTime = 0;
   amps_handle   message = 0;
@@ -1967,10 +2383,11 @@ void* amps_tcps_threaded_reader(void* userData)
     {
       goto cleanup;
     }
+    now = amps_now();
     if (me->idleTimeMillis > 0 &&
-        (amps_now() - lastIdleTime) > me->idleTimeMillis)
+        (now - lastIdleTime) > me->idleTimeMillis)
     {
-      lastIdleTime = amps_now();
+      lastIdleTime = now;
 #ifndef _WIN32
       cancelState = 0;
       pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
@@ -1981,12 +2398,77 @@ void* amps_tcps_threaded_reader(void* userData)
       pthread_setcancelstate(cancelState, &unusedCancelState);
 #endif
     }
+    if (me->sendBatchTimeoutMillis
+        && (now - me->lastSendTime) > me->sendBatchTimeoutMillis
+        && me->sendCacheOffset)
+    {
+      AMPS_LOCK(&me->sendLock);
+      int bytesSent = amps_tcps_send_cache(me);
+      AMPS_UNLOCK(&me->sendLock);
+      if (bytesSent < 0)
+      {
+        if (fd != AMPS_INVALID_SOCKET && !me->disconnecting)
+        {
+          shutdown(fd, SHUT_RDWR);
+          amps_tcps_handle_disconnect(me, connectionVersion);
+        }
+        goto cleanup;
+      }
+    }
 
     received = amps_tcps_locked_read(me, (char*) readPoint,
                                      (int)(end - readPoint), &errorcode);
 
+    if (received == AMPS_ZLIB_WANT_READ)
+    {
+      continue;
+    }
+    else if (received == AMPS_ZLIB_NEED_SPACE)
+    {
+      // Increase the buffer size for inflate resuults
+      size_t newLength = 2 * (size_t)(end - buffer);
+      newBuffer = (unsigned char*)malloc( newLength );
+      if (newBuffer == NULL)
+      {
+        /* stream broken */
+        amps_tcps_handle_stream_corruption(me, connectionVersion);
+        goto cleanup;
+      }
+      memcpy(newBuffer, buffer, (size_t)(end - buffer));
+      readPoint = newBuffer + (size_t)((unsigned const char*)readPoint - buffer);
+      if (me->useZlib
+          &&!me->disconnecting
+          && !me->destroying)
+      {
+        me->pInflater->next_out = (const char*)newBuffer
+                                  + (size_t)((unsigned const char*)me->pInflater->next_out - buffer);
+      }
+      parsePoint = newBuffer + (size_t)(parsePoint - buffer);
+#ifndef _WIN32
+      cancelState = 0;
+      pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
+#endif
+      free(buffer);
+      buffer = newBuffer;
+#ifndef _WIN32
+      pthread_setcancelstate(cancelState, &unusedCancelState);
+#endif
+      end = newBuffer + newLength;
+      assert(readPoint >= parsePoint);
+      continue;
+    }
+
     if (received > 0)
     {
+      if (me->useZlib
+          &&!me->disconnecting
+          && !me->destroying)
+      {
+        // Because the total length received could have been returned by
+        // multiple calls, get the full size.
+        received = (int)(me->pInflater->next_out - (const char*)readPoint);
+        me->pInflater->next_out = NULL;
+      }
       /* Call filter function without the 4 byte size */
       me->filterFunction(readPoint, (size_t)received, 1, me->filterUserData);
       readPoint += received;
@@ -2091,6 +2573,14 @@ void* amps_tcps_threaded_reader(void* userData)
         memcpy(newBuffer, parsePoint, (size_t)(readPoint - parsePoint));
         readPoint = newBuffer + (readPoint - parsePoint);
         parsePoint = newBuffer;
+        // Keep next_out valid as well
+        if (!me->disconnecting
+            && !me->destroying
+            && me->pInflater
+            && me->pInflater->next_out)
+        {
+          me->pInflater->next_out = (const char*)newBuffer + (size_t)((unsigned const char*)me->pInflater->next_out - buffer);
+        }
 #ifndef _WIN32
         cancelState = 0;
         pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancelState);
@@ -2111,7 +2601,9 @@ void* amps_tcps_threaded_reader(void* userData)
         break;
       }
     } /*while(readPoint >= parsePoint + 4) */
-    if (readPoint > parsePoint)
+    // If we used some bytes, but not all, and can make room, do so
+    if (readPoint > parsePoint
+        && parsePoint > buffer)
     {
       memmove(buffer, parsePoint, (size_t)(readPoint - parsePoint));
     }
@@ -2253,3 +2745,12 @@ amps_tcps_set_idle_time(amps_handle transport, int millis)
   }
   return amps_tcps_update_read_timeout(me);
 }
+
+void
+amps_tcps_set_batch_send(amps_handle transport, amps_uint64_t batchSizeBytes, amps_uint64_t timeout)
+{
+  amps_tcps_t* me = (amps_tcps_t*)transport;
+  me->sendBatchSizeBytes = batchSizeBytes;
+  me->sendBatchTimeoutMillis = timeout;
+  amps_tcps_update_read_timeout(me);
+}
