@@ -25,6 +25,9 @@
 
 #define PY_SSIZE_T_CLEAN 1
 #include <Python.h>
+#ifdef _WIN32
+#pragma warning(disable: 4996)
+#endif
 #include <amps/ampsplusplus.hpp>
 #include <ampspy_types.hpp>
 #include <ampspy_defs.hpp>
@@ -346,19 +349,25 @@ namespace ampspy
         return false;
       }
       PyObject* exception = PyObject_CallObject(ampspy::exc::AMPSException, ctor_args);
+      Py_DECREF(ctor_args);
       if (exception == NULL)
       {
         ampspy::unhandled_exception();
         return false;
       }
-      PyObject* result = PyObject_Call(_impl, Py_BuildValue("(OkO)", message, (unsigned long)requestedAckTypes_, exception), (PyObject*)NULL);
+      PyObject* args = Py_BuildValue("(OkO)", message, (unsigned long)requestedAckTypes_, exception);
+      PyObject* result = PyObject_Call(_impl, args, (PyObject*)NULL);
+      Py_DECREF(args);
+      Py_DECREF(exception);
       if (result == NULL)
       {
         ampspy::unhandled_exception();
         return false;
       }
+      bool retVal = PyObject_IsTrue(result) != 0;
+      Py_DECREF(result);
       exc::throwError();
-      return PyObject_IsTrue(result) != 0;
+      return retVal;
     }
 
   private:
