@@ -732,33 +732,37 @@ failure:
       pyFd = PyLong_FromLong((long)fd_);
     }
     // We'll only use keyword arguments
-    PyObject* args = PyTuple_New(1);
+    PyObject* args = PyTuple_New(0);
+    PyObject* kwargs = PyDict_New();
+    PyObject* pyArg = NULL;
     switch (pAddrInfo_->ai_family)
     {
     case AF_INET:
     {
-      PyTuple_SetItem(args, 0, PyObject_GetAttrString(_ampspy_socket_module, "AF_INET"));
+      pyArg = PyObject_GetAttrString(_ampspy_socket_module, "AF_INET");
     }
     break;
     case AF_INET6:
     {
-      PyTuple_SetItem(args, 0, PyObject_GetAttrString(_ampspy_socket_module, "AF_INET6"));
+      pyArg = PyObject_GetAttrString(_ampspy_socket_module, "AF_INET6");
     }
     break;
     case AF_UNIX:
     {
-      PyTuple_SetItem(args, 0, PyObject_GetAttrString(_ampspy_socket_module, "AF_UNIX"));
+      pyArg = PyObject_GetAttrString(_ampspy_socket_module, "AF_UNIX");
     }
     break;
     default:
     {
-      PyTuple_SetItem(args, 0, PyLong_FromLong((long)pAddrInfo_->ai_family));
+      pyArg = PyLong_FromLong((long)pAddrInfo_->ai_family);
     }
     }
-    PyObject* kwargs = PyDict_New();
+    PyDict_SetItemString(kwargs, "family", pyArg);
+    PyObject* pyProto = NULL;
     if (pAddrInfo_->ai_protocol)
     {
-      PyDict_SetItemString(kwargs, "proto", PyLong_FromLong(pAddrInfo_->ai_protocol));
+      pyProto = PyLong_FromLong((long)pAddrInfo_->ai_protocol);
+      PyDict_SetItemString(kwargs, "proto", pyProto);
     }
     PyDict_SetItemString(kwargs, "fileno", pyFd);
     PyObject* pySocket = PyObject_Call(pySockType, args, kwargs);
@@ -772,14 +776,15 @@ failure:
       {
         amps_tcps_set_external_error(me_, "Failed to create Python socket");
       }
+      Py_XDECREF(pyProto);
       Py_DECREF(pySockType);
       Py_DECREF(args);
+      Py_XDECREF(pyArg);
       Py_DECREF(kwargs);
       return _ampspy_get_socket_and_ssl(fd_, address_, port_, pAddrInfo_, hostname_, me_);
     }
-    Py_DECREF(args);
-    Py_DECREF(kwargs);
     PyObject* pySSLSocket = NULL;
+    PyObject* pyWrapSocket = PyString_FromString("wrap_socket");
     if (hostname_)
     {
       // We're using sni, so we need to set a few things
@@ -791,7 +796,7 @@ failure:
       Py_XDECREF(pyVerifyMode);
 
       pySSLSocket = PyObject_CallMethodObjArgs(_ampspy_ssl_context,
-                                               PyString_FromString("wrap_socket"),
+                                               pyWrapSocket,
                                                pySocket, Py_False, Py_True,
                                                Py_True,
                                                PyString_FromString(hostname_),
@@ -800,9 +805,10 @@ failure:
     else
     {
       pySSLSocket = PyObject_CallMethodObjArgs(_ampspy_ssl_context,
-                                               PyString_FromString("wrap_socket"),
+                                               pyWrapSocket,
                                                pySocket, Py_False, NULL);
     }
+    Py_DECREF(pyWrapSocket);
     if (!pySSLSocket)
     {
       if (PyErr_Occurred())
@@ -815,6 +821,10 @@ failure:
       }
       Py_DECREF(pySockType);
       Py_DECREF(pySocket);
+      Py_XDECREF(pyProto);
+      Py_DECREF(args);
+      Py_XDECREF(pyArg);
+      Py_DECREF(kwargs);
       return _ampspy_get_socket_and_ssl(fd_, address_, port_, pAddrInfo_, hostname_, me_);
     }
 
@@ -846,17 +856,13 @@ failure:
       }
 
       // We're not going to use these
-      Py_DECREF(pySockType);
-      Py_DECREF(pySocket);
       Py_DECREF(pySSLSocket);
-      // Ignore errors here, just being sure the socket is released
-#ifdef _WIN32
-      shutdown(fd_, SD_BOTH);
-      closesocket(fd_);
-#else
-      shutdown(fd_, SHUT_RDWR);
-      close(fd_);
-#endif
+      Py_DECREF(pySocket);
+      Py_DECREF(pySockType);
+      Py_XDECREF(pyProto);
+      Py_DECREF(args);
+      Py_XDECREF(pyArg);
+      Py_DECREF(kwargs);
       return NULL;
     }
     Py_DECREF(retVal);
@@ -882,6 +888,10 @@ failure:
 
     AMPS::Lock<AMPS::Mutex> guard(_ampspy_ssl_map_lock);
     _ampspy_ssl_object_map[ssl] = std::make_pair(pySSLSocket, pySocket);
+    Py_XDECREF(pyProto);
+    Py_DECREF(args);
+    Py_XDECREF(pyArg);
+    Py_DECREF(kwargs);
     Py_DECREF(pySSLSocketObj);
     Py_DECREF(pySockType);
     return ssl;
@@ -892,6 +902,7 @@ failure:
   {
     if (!pSSL_)
     {
+      assert(fd_ == AMPS_INVALID_SOCKET);
       if (fd_ != AMPS_INVALID_SOCKET)
       {
 #ifdef _WIN32
@@ -921,13 +932,17 @@ failure:
       LOCKGIL;
       // Shutdown everything, then close the connection
       PyObject* arg = PyObject_GetAttrString(_ampspy_socket_module, "SHUT_RDWR");
-      PyObject_CallMethodObjArgs(pySSLSocket, PyString_FromString("shutdown"), arg, NULL);
-      PyObject_CallMethodObjArgs(pySSLSocket, PyString_FromString("_real_close"), NULL);
+      PyObject* shutdownStr = PyString_FromString("shutdown");
+      PyObject* closeStr = PyString_FromString("close");
+      PyObject_CallMethodObjArgs(pySSLSocket, shutdownStr, arg, NULL);
+      PyObject_CallMethodObjArgs(pySSLSocket, closeStr, NULL);
       PyObject_SetAttrString(pySSLSocket, "_closed", Py_True);
       PyObject_SetAttrString(pySSLSocket, "connected", Py_False);
       Py_XDECREF(arg);
-      Py_DECREF(pySSLSocket);
-      Py_DECREF(pySocket);
+      Py_DECREF(shutdownStr);
+      Py_DECREF(closeStr);
+      Py_XDECREF(pySSLSocket);
+      Py_XDECREF(pySocket);
     }
     else if (!_ampspy_socket_module)
     {
